from pyliftover import LiftOver

CHROM_LIST = {}
CHROM_LEN = {}
CHROM_TYPE = {}
CHROM_TYPE_LIST = ['autosome', 'sex', 'mt', 'unlocalized', 'unplaced', 'haplotype', 'hla', 'ebv', 'decoy']
REF_SEQ_VERSION_LIST = ['b37', 'hg19', 'b37d5', 'b38', 'hg38', 'b38d']
MAIN_CHROM_LIST = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22", "X", "Y", "M"]

codon = {}
codon['UUU'] = "F"
codon['UUC'] = "F"
codon['UUA'] = "L"
codon['UUG'] = "L"
codon['CUU'] = "L"
codon['CUC'] = "L"
codon['CUA'] = "L"
codon['CUG'] = "L"
codon['AUU'] = "I"
codon['AUC'] = "I"
codon['AUA'] = "I"
codon['AUG'] = "M"
codon['GUU'] = "V"
codon['GUC'] = "V"
codon['GUA'] = "V"
codon['GUG'] = "V"
codon['UCU'] = "S"
codon['UCC'] = "S"
codon['UCA'] = "S"
codon['UCG'] = "S"
codon['CCU'] = "P"
codon['CCC'] = "P"
codon['CCA'] = "P"
codon['CCG'] = "P"
codon['ACU'] = "T"
codon['ACC'] = "T"
codon['ACA'] = "T"
codon['ACG'] = "T"
codon['GCU'] = "A"
codon['GCC'] = "A"
codon['GCA'] = "A"
codon['GCG'] = "A"
codon['UAU'] = "Y"
codon['UAC'] = "Y"
codon['UAA'] = "*"
codon['UAG'] = "*"
codon['CAU'] = "H"
codon['CAC'] = "H"
codon['CAA'] = "Q"
codon['CAG'] = "Q"
codon['AAU'] = "N"
codon['AAC'] = "N"
codon['AAA'] = "K"
codon['AAG'] = "K"
codon['GAU'] = "D"
codon['GAC'] = "D"
codon['GAA'] = "E"
codon['GAG'] = "E"
codon['UGU'] = "C"
codon['UGC'] = "C"
codon['UGA'] = "*"
codon['UGG'] = "W"
codon['CGU'] = "R"
codon['CGC'] = "R"
codon['CGA'] = "R"
codon['CGG'] = "R"
codon['AGU'] = "S"
codon['AGC'] = "S"
codon['AGA'] = "R"
codon['AGG'] = "R"
codon['GGU'] = "G"
codon['GGC'] = "G"
codon['GGA'] = "G"
codon['GGG'] = "G"

complementary = {}
complementary['A'] = 'T'
complementary['T'] = 'A'
complementary['G'] = 'C'
complementary['C'] = 'G'

BASE = ["A", "T", "G", "C"]
COMPBASE = {'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G'}

# SIGNATURE
SINGLEBASE_PATTERN6 = ['C>A', 'C>G', 'C>T', 'T>A', 'T>C', 'T>G']
SINGLEBASE_PATTERN = ['C>A', 'G>T', 'C>G', 'G>C', 'C>T', 'G>A', 'T>A', 'A>T', 'T>C', 'A>G', 'T>G', 'A>C']
SINGLEBASE_COMPLEMENTARY = {'G>T': 'C>A', 'C>A': 'C>A', 'C>G': 'C>G', 'G>C': 'C>G', 'C>T': 'C>T', 'G>A': 'C>T',
                            'T>A': 'T>A', 'A>T': 'T>A', 'T>C': 'T>C', 'A>G': 'T>C', 'T>C': 'T>C', 'A>G': 'T>C', 'T>G': 'T>G', 'A>C': 'T>G'}

EXOM_SEQ_COVERED_REGION_NimbleGen_SeqCap = 64190747
EXOM_SEQ_COVERED_REGION = EXOM_SEQ_COVERED_REGION_NimbleGen_SeqCap

REF_FASTA = {}
REF_FASTA['b37d5'] = "/home/mk446/BiO/Install/GATK-bundle/2.8/b37/human_g1k_v37_decoy.fasta"
REF_FASTA['hg38'] = "/home/mk446/BiO/Install/GATK-bundle/2.8/hg38/Homo_sapiens_assembly38.fasta"
REF_FASTA['hg19'] = "/home/mk446/BiO/Install/GATK-bundle/2.8/hg19/ucsc.hg19.fasta"
REF_FASTA['b38d'] = "/home/mk446/BiO/Install/GATK-bundle/2.8/b38/GRCh38_full_analysis_set_plus_decoy_hla.fa"


def load_liftover(chainfile = ""):
    if chainfile == "":
        lo = LiftOver('hg38', 'hg19')
    else:
        lo = LiftOver(chainfile)
    return lo

def convert_coordinate(lo_hg38_hg19, chrom, pos):
    if 'chr' not in chrom:
        chrom = 'chr' + chrom
    if type(pos) != int:
        pos = int(pos)
    return lo_hg38_hg19.convert_coordinate(chrom, pos)



def load_refseq_info(seq_ver=''):
    global CHROM_LIST, CHROM_LEN, CHROM_TYPE, REF_SEQ_VERSION_LIST
    path = '/home/mk446/bin/python_lib/'
    if seq_ver == '':
        for seq_ver in REF_SEQ_VERSION_LIST:
            load_refseq_info(seq_ver)
    else:
        CHROM_LIST[seq_ver] = []
        CHROM_LEN[seq_ver] = {}
        CHROM_TYPE[seq_ver] = {}
        for line in open(path+'ref_'+seq_ver+'.txt'):
            arr = line.strip().split('\t')
            if arr[0] != "CHROM" and arr[0] != "":
                CHROM_LIST[seq_ver].append(arr[0])
                CHROM_LEN[seq_ver][arr[0]] = int(arr[1])
                CHROM_TYPE[seq_ver][arr[0]] = arr[2]


# load_refseq_info('b37d5')
# load_refseq_info('b38d')


def get_split_region(chunksize=1000000, seqver='b38d'):
    load_refseq_info(seqver)
    rst = []
    k2 = 1
    for chrom in CHROM_LIST[seqver]:
        if len(chrom) < 3:
            no_split = int(1.0 * CHROM_LEN[seqver][chrom] / chunksize)
            for k in range(1, no_split + 2):
                spos = chunksize * (k - 1) + 1
                epos = chunksize * (k)
                if epos > CHROM_LEN[seqver][chrom]:
                    epos = CHROM_LEN[seqver][chrom]
                rst.append((chrom, spos, epos, k2))
                k2 += 1
    return rst


def codeAA(dnaseq3):
    global codon
    aa = ""
    if len(dnaseq3) == 3:
        aa = codon[dnaseq3.upper().replace('T', 'U')]
    return aa


def complementarySeq(seq, direction='+'):
    global complementary
    cseq = ""
    if direction == '-':
        for i in range(0, len(seq)):
            cseq += complementary[seq[i].upper()]
    else:
        for i in range(len(seq)-1, -1, -1):
            cseq += complementary[seq[i].upper()]

    return cseq

# DNA to Protein


def translate(dnaseq, orient="+"):
    global codon
    p = ""
    if orient == "-":
        r_dnaseq = dnaseq[::-1]

    for i in range(0, len(dnaseq), 3):
        if orient == "-":
            p += codeAA(complementarySeq(r_dnaseq[i:(i+3)]))
        else:
            p += codeAA(dnaseq[i:(i+3)])

    return p


# DNA to Protein
def translateAll(dnaseq):
    global codon
    p = {}
    p['0'] = ""
    p['1'] = ""
    p['2'] = ""
    p['-0'] = ""
    p['-1'] = ""
    p['-2'] = ""
    r_dnaseq = dnaseq[::-1]
    for i in range(0, len(dnaseq), 3):
        p['0'] += codeAA(dnaseq[i:(i+3)])
        p['1'] += codeAA(dnaseq[(i+1):(i+1+3)])
        p['2'] += codeAA(dnaseq[(i+2):(i+2+3)])
        p['-0'] += codeAA(complementarySeq(r_dnaseq[i:(i+3)]))
        p['-1'] += codeAA(complementarySeq(r_dnaseq[(i+1):(i+1+3)]))
        p['-2'] += codeAA(complementarySeq(r_dnaseq[(i+2):(i+2+3)]))

    return p


def get_chroSeqLen(seqlen_file="/ms2/refseq/hg19/chr/chr_len.txt"):
    m = {}
    for line in open(seqlen_file):
        arr = line.strip().split("\t")
        m[arr[0]] = int(arr[1])
    return m

# OUTDATED.... USE get_refseq()
# def get_seq(chrom, spos, epos, strand='+', seq_path="/home/mk446/BiO/Install/GATK-bundle/2.8/b37"):
#    fasta = seq_path + "/human_g1k_v37_decoy." + chrom + ".fasta"
#    #fasta = maestro.FASTA['b37_chrom'].replace("#CHROM#",chrom)
#    f = open(fasta, "r")
#    f.seek(spos-1) ### start 1
#    rst = f.read(epos-spos+1)
#    if strand == '-':
#        rst = complementarySeq(rst)
#    f.close()
#    return rst

# UCSC hg19 - outdated


def get_dnaseq(chro, spos, epos, orient="+", seq_path="/ms2/refseq/hg19/chr/", seqlen_file="/ms2/refseq/hg19/chr/chr_len.txt", report_pos=False):
    seq_file = seq_path + "chr" + chro + ".fa"

    if orient == "-":
        seqlen = get_chroSeqLen(seqlen_file)
        sspos = seqlen['chr'+chro] - epos
        eepos = seqlen['chr'+chro] - spos
    else:
        sspos = spos
        eepos = epos

    # print "sspos", sspos
    # print "eepos", eepos

    i = 0
    this_line_spos = 0
    sidx = 0
    eidx = 0
    seq = ""
    for line in open(seq_file):
        if i > 0:
            next_pos = this_line_spos + len(line.strip())
            if sspos >= this_line_spos and sspos < next_pos:
                sidx = sspos-this_line_spos
            if sidx > 0:
                seq += line.strip()
            if eepos >= this_line_spos and eepos < next_pos:
                eidx = eepos-this_line_spos
                break
            this_line_spos += len(line.strip())
        i += 1

    dnaseq = seq[sidx:(eepos-sspos)+sidx]
    # print seq, epos-sspos+1, len(dnaseq)
    # print sidx, eidx
    if (report_pos):
        return (dnaseq, sspos, eepos)
    else:
        return dnaseq


def get_base_info_from_bam(bam, chrom, pos1, ref="", alt=""):
    import pysam
    pysamAF = pysam.AlignmentFile(bam, "rb")
    return get_base_info_from_pysamAF(pysamAF, chrom, pos1, ref="", alt="")


def get_base_info_from_pysamAF(pysamAF, chrom, pos1, ref="", alt=""):
    pos = pos1-1
    #pos = pos1+1
    info = {}
    for x in pysamAF.pileup(chrom, pos-5, pos+5, min_base_quality=0):
        if x.reference_pos == pos:
            # print x.nsegments, x.reference_pos, len(x.pileups)
            info['total_depth'] = len(x.pileups)
            info['chrom'] = chrom
            info['pos'] = x.reference_pos
            info['base_info'] = []
            base_stat = get_base_info_from_bam_init_base_stat()
            base_stat_mapq20 = get_base_info_from_bam_init_base_stat()
            base_stat_baseq20 = get_base_info_from_bam_init_base_stat()
            base_stat_mapq20baseq20 = get_base_info_from_bam_init_base_stat()
            for pr in x.pileups:
                a = pr.alignment
                # print pr.query_position, a.query_sequence[pr.query_position], a.query_qualities[pr.query_position], a.mapq, a.is_reverse
                try:
                    base = {}
                    base['pos_in_read'] = pr.query_position
                    base['base'] = a.query_sequence[pr.query_position]
                    base['base_qual'] = a.query_qualities[pr.query_position]
                    base['MAPQ'] = a.mapq
                    base['is_reverse'] = a.is_reverse
                    info['base_info'].append(base)

                    base_stat[a.query_sequence[pr.query_position]] += 1
                    if base['MAPQ'] >= 20:
                        base_stat_mapq20[a.query_sequence[pr.query_position]] += 1
                    if base['base_qual'] >= 20:
                        base_stat_baseq20[a.query_sequence[pr.query_position]] += 1
                    if base['base_qual'] >= 20 and base['MAPQ'] >= 20:
                        base_stat_mapq20baseq20[a.query_sequence[pr.query_position]] += 1

                except TypeError:
                    pass
            info['base_stat'] = add_af_dp_from_base_stat(base_stat, ref, alt)
            info['base_stat_mapq20'] = add_af_dp_from_base_stat(base_stat_mapq20, ref, alt)
            info['base_stat_baseq20'] = add_af_dp_from_base_stat(base_stat_baseq20, ref, alt)
            info['base_stat_mapq20baseq20'] = add_af_dp_from_base_stat(base_stat_mapq20baseq20, ref, alt)
    return info


def get_baserange_info_from_bam(bam, chrom, spos, epos, flag_read_info=False):
    import pysam
    infolist = {}
    samfile = pysam.AlignmentFile(bam, "rb")
    ref = ''
    alt = ''
    for x in samfile.pileup(chrom, spos-50, epos+50, min_base_quality=0):
        if x.reference_pos >= spos-1 and x.reference_pos <= epos+1:
            # print x.nsegments, x.reference_pos, len(x.pileups)
            info = {}
            info['total_depth'] = len(x.pileups)
            info['chrom'] = chrom
            info['pos'] = x.reference_pos
            info['base_info'] = []
            base_stat = get_base_info_from_bam_init_base_stat()
            base_stat_mapq20 = get_base_info_from_bam_init_base_stat()
            base_stat_baseq20 = get_base_info_from_bam_init_base_stat()
            base_stat_mapq20baseq20 = get_base_info_from_bam_init_base_stat()
            for pr in x.pileups:
                a = pr.alignment
                # print pr.query_position, a.query_sequence[pr.query_position], a.query_qualities[pr.query_position], a.mapq, a.is_reverse
                try:
                    base = {}
                    base['pos_in_read'] = pr.query_position
                    base['base'] = a.query_sequence[pr.query_position]
                    base['base_qual'] = a.query_qualities[pr.query_position]
                    base['MAPQ'] = a.mapq
                    base['is_reverse'] = a.is_reverse
                    if flag_read_info:
                        info['base_info'].append(base)

                    base_stat[a.query_sequence[pr.query_position]] += 1
                    if base['MAPQ'] >= 20:
                        base_stat_mapq20[a.query_sequence[pr.query_position]] += 1
                    if base['base_qual'] >= 20:
                        base_stat_baseq20[a.query_sequence[pr.query_position]] += 1
                    if base['base_qual'] >= 20 and base['MAPQ'] >= 20:
                        base_stat_mapq20baseq20[a.query_sequence[pr.query_position]] += 1

                except TypeError:
                    pass
            info['base_stat'] = add_af_dp_from_base_stat(base_stat, ref, alt)
            info['base_stat_mapq20'] = add_af_dp_from_base_stat(base_stat_mapq20, ref, alt)
            info['base_stat_baseq20'] = add_af_dp_from_base_stat(base_stat_baseq20, ref, alt)
            info['base_stat_mapq20baseq20'] = add_af_dp_from_base_stat(base_stat_mapq20baseq20, ref, alt)
            infolist[x.reference_pos+1] = info
    return infolist


def add_af_dp_from_base_stat(base_stat, ref, alt):
    if ref != "":
        base_stat['DP'] = base_stat['A']+base_stat['T']+base_stat['G']+base_stat['C']+base_stat['N']
        if base_stat['DP'] > 0:
            base_stat['AF'] = base_stat[alt]/base_stat['DP']
        else:
            base_stat['AF'] = 0
        base_stat['AD'] = base_stat[alt]
    return base_stat


def get_base_info_from_bam_init_base_stat():
    base_stat = {}
    base_stat['A'] = 0
    base_stat['T'] = 0
    base_stat['G'] = 0
    base_stat['C'] = 0
    base_stat['N'] = 0
    return base_stat

## samf : samf = pysam.AlignmentFile(bam, "rb")
## chrom : arr[0]
## pos = int(arr[1])
## ref = arr[4]
## alt = arr[5]
# return giv : odds ratio


def get_giv(samf, chrom, pos, ref, alt):
    d = {}
    cnt_reads = load_reads(samf, chrom, pos)
    cnt_r1_ref = get_value(cnt_reads['r1'], ref)
    cnt_r1_alt = get_value(cnt_reads['r1'], alt)
    cnt_r2_ref = get_value(cnt_reads['r2'], ref)
    cnt_r2_alt = get_value(cnt_reads['r2'], alt)
    if cnt_r1_ref == 0:
        iv_r1 = 1.0
    else:
        iv_r1 = (1.0*cnt_r1_alt/cnt_r1_ref)
    if cnt_r2_ref == 0:
        iv_r2 = 1.0
    else:
        iv_r2 = (1.0*cnt_r2_alt/cnt_r2_ref)

    if iv_r2 == 0:
        giv = 1.0
    else:
        giv = iv_r1 / iv_r2

    d['cnt_reads'] = cnt_reads
    d['cnt_r1_ref'] = cnt_r1_ref
    d['cnt_r1_alt'] = cnt_r1_alt
    d['cnt_r2_ref'] = cnt_r2_ref
    d['cnt_r2_alt'] = cnt_r2_alt
    d['iv_r1'] = iv_r1
    d['iv_r2'] = iv_r2
    d['giv'] = giv
    return d


def get_value(m, k1):
    try:
        v1 = m[k1]
    except KeyError:
        v1 = 0
    return v1


def load_reads(samf, chrom, pos_plus_1):
    pos = pos_plus_1 - 1
    m = {}
    m['r1'] = {}
    m['r2'] = {}
    for x in samf.pileup(chrom, pos-1, pos+1):
        if x.reference_pos == pos:
            # print x.nsegments, x.reference_pos, len(x.pileups)
            for pr in x.pileups:
                a = pr.alignment
                if pr.query_position != None:
                    base = a.query_sequence[pr.query_position]
                    # print pr.query_position, a.query_sequence[pr.query_position], a.query_sequence, a.is_reverse, a.is_read1, a.is_read2
                    # print pr.query_position, a.query_sequence[pr.query_position], a.query_qualities[pr.query_position], a.mapq, a.is_reverse
                    if a.mapq >= 20 and not a.is_duplicate and not a.is_unmapped:
                        if a.is_read1:
                            try:
                                m['r1'][base] += 1
                            except KeyError:
                                m['r1'][base] = 1
                        elif a.is_read2:
                            try:
                                m['r2'][base] += 1
                            except KeyError:
                                m['r2'][base] = 1

    return m


def get_refseq(chrom, spos, epos, seqver="b38d"):
    from pyfasta import Fasta
    #refseq = seq_path + "/human_g1k_v37_decoy.fasta"
    spos = spos - 1
    refseq = REF_FASTA[seqver]
    f = Fasta(refseq)
    #print (sorted(f.keys()))
    chromlist = list(f.keys())
    #print (chromlist[:30])
    fasta_chrom = ""
    for c1 in chromlist:
        if c1.split(' ')[0] == chrom:
            fasta_chrom = c1
            break
    #print (fasta_chrom)
    #fasta_chrom = chrom + ' dna:chromosome chromosome:GRCh37:'+chrom+':1:'+str(CHROM_LEN['b37d5'][chrom])+':1'
    refseq = f[fasta_chrom][spos:epos+1]
    return refseq


############################################
# SEQ QC
############################################
def get_qcfile_info(fname, qctool):
    if qctool == 'stat':
        info = get_bamstat_info(fname)
    if qctool == 'cmm':
        info = get_cmm_info(fname)
    if qctool == 'hist':
        info = get_hist_info(fname)
    return info


def get_cmm_info(fname):
    # fname : .cmm.alignment_summary_metrics file
    info = {}

    for line in open(fname):
        if line[:len('FIRST_OF_PAIR')] == 'FIRST_OF_PAIR':
            arr = line.split('\t')
            info['first_of_pair'] = arr[1].strip()
        if line[:len('SECOND_OF_PAIR')] == 'SECOND_OF_PAIR':
            arr = line.split('\t')
            info['second_of_pair'] = arr[1].strip()
        if line[:len('PAIR')] == 'PAIR':
            arr = line.split('\t')
            info['total_reads'] = arr[1].strip()
            info['mean_read_length'] = arr[15].strip()

    info['genome_coverage'] = str(round(int(info['total_reads']) * int(info['mean_read_length']) / CHROM_TOTAL_LEN, 3))
    info['exom_coverage'] = str(round(int(info['total_reads']) * int(info['mean_read_length']) / EXOM_SEQ_COVERED_REGION, 3))

    insert_size_metrics_header = []
    flag_metrics = False
    flag_size_dist = False
    info['insert_size_dist'] = ""
    for line in open(fname.replace('.cmm.alignment_summary_metrics', '.cmm.insert_size_metrics')):
        if flag_metrics:
            arr = line.split('\t')
            arr[-1] = arr[-1].strip()
            for k in range(len(arr)):
                info[insert_size_metrics_header[k]] = arr[k].strip()
            flag_metrics = False
            break
        if flag_size_dist:
            line = line.strip()
            if line(line) > 2:
                info['insert_size_dist'] += line+'\n'

        if line[:len('MEDIAN_INSERT_SIZE')] == 'MEDIAN_INSERT_SIZE':
            insert_size_metrics_header = line.split('\t')
            insert_size_metrics_header[-1] = insert_size_metrics_header[-1].strip()
            flag_metrics = True

        if line[:len('insert_size')] == 'insert_size':
            flag_size_dist = True

    ### TODO .cmm.quality_by_cycle_metrics
    ### TODO .cmm.quality_distribution_metrics
    ### TODO .cmm.base_distribution_by_cycle_metrics
    ### TODO .cmm.quality_distribution_metrics
    return info


def get_hist_info(fname):
    info = {}
    read_depth10 = 0
    read_depth30 = 0
    read_depth50 = 0
    read_depth70 = 0
    read_depth100 = 0
    read_depth200 = 0
    for line in open(fname):
        if line[:len('genome')] == 'genome':

            arr = line.strip().split('\t')
            read_depth = int(arr[1])
            covered_region = arr[2]
            # print (arr,read_depth,read_depth)
            if read_depth == 0:
                info['uncovered_region'] = covered_region
            elif read_depth <= 400:
                info['covered_region (dp='+str(read_depth)+')'] = covered_region
            if read_depth > 10:
                read_depth10 += int(covered_region)
            if read_depth > 30:
                read_depth30 += int(covered_region)
            if read_depth > 50:
                read_depth50 += int(covered_region)
            if read_depth > 70:
                read_depth70 += int(covered_region)
            if read_depth > 100:
                read_depth100 += int(covered_region)
            if read_depth > 200:
                read_depth200 += int(covered_region)
    info['covered_region (dp>10)'] = str(read_depth10)
    info['covered_region (dp>30)'] = str(read_depth30)
    info['covered_region (dp>50)'] = str(read_depth50)
    info['covered_region (dp>70)'] = str(read_depth70)
    info['covered_region (dp>100)'] = str(read_depth100)
    info['covered_region (dp>200)'] = str(read_depth200)

    return info


def get_bamstat_info(fname):
    info = {}
    flag_all_chrom = False
    flag_only_main = False
    flag_xy = False
    seqversion = "b37"

    for line in open(fname):
        if '===ALL CHROM==' in line:
            flag_all_chrom = True

        if '=ONLY MAIN CHROM (1~22,X,Y,MT)==' in line:
            flag_only_main = True
            flag_all_chrom = False

        if '==X,Y CHROM==' in line:
            flag_xy = True
            flag_only_main = False
            flag_all_chrom = False

        if "READ_LEN" in line:
            info['read_length'] = line.strip().replace('READ_LEN:', '').strip()

        if not flag_only_main and not flag_all_chrom and not flag_xy:
            arr = line.split('\t')
            # print (arr)
            if arr[0] != "CHROM":
                info[arr[0] + '_chrom_mapped_reads'] = arr[2].replace(',', '')
                info[arr[0] + '_chrom_unmapped_reads'] = arr[3].replace(',', '')
                info[arr[0] + '_chrom_total_reads'] = arr[4].replace(',', '')
                if arr[0] == "1" and arr[1] == "248,956,422":
                    seqversion = "b38"
                if arr[0] == "1" and arr[1] == "249,250,621":
                    seqversion = "b37"
                if arr[0] == "hs37d5":
                    seqversion = "b37d5"
                if arr[0] == "6_apd_hap1":
                    seqversion = "hg19"
                if arr[0] == "19_KI270938v1_alt":
                    seqversion = "hg38"
                if arr[0] == "Un_JTFH01001998v1_decoy":
                    seqversion = "b38d"

                # if line[0] == "X":
                #     info['X_chrom_total_reads'] = arr[4].replace(',','')
                # if line[0] == "Y":
                #     info['Y_chrom_total_reads'] = arr[4].replace(',','')

        if flag_all_chrom and "COVERAGE:" in line:
            info['all_chrom_coverage'] = line.strip().replace('COVERAGE:', '').strip()
        if flag_all_chrom and "TOTAL READS:" in line:
            info['all_chrom_total_reads'] = line.strip().replace('TOTAL READS:', '').replace(',', '').strip()
        if flag_all_chrom and "MAPPED:" in line and not "UNMAPPED:" in line:
            info['all_chrom_mapped_reads'] = line.strip().replace('MAPPED:', '').replace(',', '').strip()
        if flag_all_chrom and "UNMAPPED:" in line:
            info['all_chrom_unmapped_reads'] = line.strip().replace('UNMAPPED:', '').replace(',', '').strip()

        if flag_only_main and "COVERAGE:" in line:
            info['main_chrom_coverage'] = line.strip().replace('COVERAGE:', '').split('(')[0].strip()
        if flag_only_main and "TOTAL READS:" in line:
            info['main_chrom_total_reads'] = line.strip().replace('TOTAL READS:', '').replace(',', '').split('(')[0].strip()
        if flag_only_main and "MAPPED:" in line and not "UNMAPPED:" in line:
            info['main_chrom_mapped_reads'] = line.strip().replace('MAPPED:', '').replace(',', '').split('(')[0].strip()
        if flag_only_main and "UNMAPPED:" in line:
            info['main_chrom_unmapped_reads'] = line.strip().replace('UNMAPPED:', '').replace(',', '').split('(')[0].strip()

        if flag_xy and "X/Y RATIO:" in line:
            info['xy_ratio'] = line.strip().replace('X/Y RATIO:', '').replace(',', '').split('(')[0].strip()
        if flag_xy and "EXT. GENDER:" in line:
            info['ext_gender'] = line.strip().replace('EXT. GENDER:', '').replace(',', '').split('(')[0].strip()

    info['all_chrom_mapped_ratio'] = str(round(int(info['all_chrom_mapped_reads'])/int(info['all_chrom_total_reads'])*100, 4))+'%'
    if int(info['main_chrom_total_reads']) > 0:
        info['main_chrom_mapped_ratio'] = str(round(int(info['main_chrom_mapped_reads'])/int(info['main_chrom_total_reads'])*100, 4))+'%'
    else:
        info['main_chrom_mapped_ratio'] = "0"
    info['main_chrom_read_ratio'] = str(round(int(info['main_chrom_total_reads'])/int(info['all_chrom_total_reads'])*100, 4))+'%'
    info['seqversion'] = seqversion
    return info



from pyfaidx import Fasta
from os import path
from .util import file_util

VEP_OPTION_CMD = "#VEP# -i #INPUTVCF# -o #OUT#"
VEP_OPTION_CMD += " --hgvs"
VEP_OPTION_CMD += " --fasta #FASTA#"
VEP_OPTION_CMD += " --assembly GRCh38"
VEP_OPTION_CMD += " --use_given_ref"
VEP_OPTION_CMD += " --offline"
VEP_OPTION_CMD += " --cache_version #CACHE_VERSION#"
VEP_OPTION_CMD += " --dir_cache #VEPCACHE#"
VEP_OPTION_CMD += " --everything"
VEP_OPTION_CMD += " --force_overwrite"
VEP_OPTION_CMD += " --vcf"
# PLUGIN
# VEP_OPTION_CMD += "--plugin LoF --plugin LoF,human_ancestor_fa:/home/mk446/BiO/Data/vep/human_ancestor.fa.gz "
VEP_OPTION_CMD += " --plugin MaxEntScan,/home/mk446/bio/mutanno/ANNOT3TOOLS/BIN/VEP_plugins-release-99/fordownload"
VEP_OPTION_CMD += " --plugin TSSDistance"
VEP_OPTION_CMD += " --dir_plugins /home/mk446/bio/mutanno/ANNOT3TOOLS/BIN/VEP_plugins-release-99"
# v99 added
VEP_OPTION_CMD += " --plugin SpliceRegion,Extended"
# VEP_OPTION_CMD += " --plugin LoF,loftee_path:/home/mk446/bio/mutanno/ANNOT3TOOLS/BIN/VEP_plugins-release-99/loftee"


class PreCalculate():
    def __init__(self, opt):
        self.opt = opt

    def getseq(self, chrom, spos, epos):
        pass

    def get_split_region(self, fastaobj, chunksize=1000000):
        chunklist = []
        k2 = 1
        for chrom in list(fastaobj.keys()):
            if len(chrom) < 6:
                no_split = int(1.0 * len(fastaobj[chrom]) / chunksize)
                for k in range(1, no_split + 2):
                    spos = chunksize * (k - 1) + 1
                    epos = chunksize * (k)
                    if epos > len(fastaobj[chrom]):
                        epos = len(fastaobj[chrom])
                    chunklist.append((chrom, spos, epos, k2))
                    k2 += 1
        return chunklist

    def save_input_vcf(self, fastaobj, inputvcf, chrom, spos, epos):
        file_util.check_dir(inputvcf)
        f = open(inputvcf, 'w')
        header = '#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tS1'
        f.write(header + '\n')
        for k in range(spos, epos + 1):
            ref = fastaobj[chrom][k - 1]
            print(ref)
            if ref != 'N' and ref.strip() != '':
                cont = chrom
                cont += '\t' + str(k)
                cont += '\t' + '.'
                cont += '\t' + ref
                cont += '\t#ALT#'
                cont += '\t' + '.'
                cont += '\t' + '.'
                cont += '\t' + '.'
                cont += '\t' + 'GT:AD:DP'
                cont += '\t' + '0/1:0,30:60'
                cont += '\n'
                for alt in ['A', 'T', 'G', 'C']:
                    if ref != alt:
                        f.write(cont.replace('#ALT#', alt))
        f.close()

    def save_runcmd_vep(self, inputvcf, chrom, spos, epos):
        cmd = ""
        if not file_util.is_exist(inputvcf):
            cmd += "mutanno precal"
            cmd += " -make_input_vcf "
            cmd += " -out " + inputvcf
            for param in ['fasta']:
                cmd += " -" + param + " " + self.opt[param]
            cmd += " -region " + chrom + ':' + str(spos) + '-' + str(epos)
            cmd += ";"
            cmd += "sleep 60;"
        out = self.get_vepout_path(inputvcf)
        vepcmd = VEP_OPTION_CMD
        vepcmd = vepcmd.replace('#VEP#', self.opt['vep'])
        vepcmd = vepcmd.replace('#INPUTVCF#', inputvcf)
        vepcmd = vepcmd.replace('#OUT#', out)
        vepcmd = vepcmd.replace('#FASTA#', path.abspath(self.opt['fasta']))
        vepcmd = vepcmd.replace('#VEPCACHE#', path.abspath(self.opt['vepcache']))
        vepcmd = vepcmd.replace('#CACHE_VERSION#', self.opt['cache_version'])
        cmd += vepcmd + ";"
        cmd += "touch " + inputvcf + ".vep.txt.done;"

        print(cmd)
        # file_util.fileSave(out, cmd, 'w')
        # proc_util.run_cmd('chmod 755 ' + out)

    def get_inputvcf_path(self, chrom, spos, epos):
        out = path.join(self.opt['out'], chrom, str(int(spos / 1000000)),
                        chrom + '_' + str(spos) + '_' + str(epos) + '.vcf')
        out = path.abspath(out)
        return out

    def get_vepout_path(self, inputvcf):
        return inputvcf + '.vep.txt'

    def mk_run_vep_script(self, chunklist):
        for (chrom, spos, epos, k2) in chunklist:
            inputvcf = self.get_inputvcf_path(chrom, spos, epos)
            self.save_runcmd_vep(inputvcf, chrom, spos, epos)
            # if k2 % 1000 == 0:
            #     print("processing..", k2, chrom, spos, inputvcf)
            # break

    # vep_type: vcf, txt
    def check_vep_result(self, vcf, vep_result, output_type='vcf'):
        print('check_vep_result')
        varmap = {}
        for line in file_util.gzopen(vcf):
            if vcf.endswith('.gz'):
                line = line.decode('UTF-8')
            if line[0] != '#':
                arr = line.split('\t')
                # vid = arr[0] + '_' + arr[1] + '_' + arr[3] + '/' + arr[4]
                # (ref, alt, pos) = vcf_util.remove_nonvariant_base(arr[3], arr[4], int(arr[1]), '-')
                pos = int(arr[1])
                ref = arr[3]
                alt = arr[4]
                if output_type == 'txt':
                    if ref[0] == alt[0]:
                        ref = ref[1:]
                        alt = alt[1:]
                        pos += 1
                vid = arr[0] + '_' + str(pos) + '_' + ref + '/' + alt
                # print(vid, ref, alt, pos)
                varmap[vid] = 0

        cont = "no. of variants in VCF: " + str(len(varmap.keys())) + '\n'
        additional_varno = {}
        mismatch_column = {}
        total_lines = 0
        colno = 0
        colidx = {}
        for line in file_util.gzopen(vep_result):
            if vep_result.endswith('.gz'):
                line = line.decode('UTF-8')
            if line[:2] == "##":
                pass
            elif line[0] == "#":
                arr = line[1:].split('\t')
                for k in range(len(arr)):
                    colidx[arr[k]] = k
                colno = len(arr)
                cont += 'no. of column in VEP: ' + str(colno) + '\n'
            elif line[0] != '#':
                arr = line.split('\t')
                if output_type == "txt":
                    vid = arr[colidx['Uploaded_variation']]
                elif output_type == "vcf":
                    vid = arr[0] + '_' + arr[1] + '_' + arr[3] + '/' + arr[4]
                if colno != len(arr):
                    mismatch_column[vid] = 1
                try:
                    varmap[vid] += 1
                except KeyError:
                    additional_varno[vid] = 1
                    print(vid)
                total_lines += 1

        miss_varno = 0
        annot_varno = 0
        for k1 in varmap.keys():
            if varmap[k1] == 0:
                miss_varno += 1
            else:
                annot_varno += 1
        cont += "total lines in VEP: " + str(total_lines) + '\n'
        cont += "no. of annotated variants in VEP: " + str(annot_varno) + '\n'
        cont += "no. of missing variants in VEP: " + str(miss_varno) + '\n'
        cont += "no. of variants with mismatching column in VEP: " + str(len(mismatch_column.keys())) + '\n'
        cont += "no. of additional variants in VEP: " + str(len(additional_varno.keys())) + '\n'
        print(cont)
        if miss_varno > 0 or len(mismatch_column.keys()) > 0:
            out = vep_result + '.error'
        else:
            out = vep_result + '.checked'
        file_util.fileSave(out, cont, 'w')
        print("Saved", out)

    def mk_shellscript_merge_vep(self, fasta, out, chunklist):
        for (chrom, spos, epos, k2) in chunklist:
            inputvcf = self.get_inputvcf_path(chrom, spos, epos)
            print(inputvcf)
        pass

    def run(self):
        chunksize = 100000
        if self.opt['fasta'] != "":
            fasta = Fasta(self.opt['fasta'], as_raw=True, sequence_always_upper=True)

        if self.opt['run_vep']:
            chunklist = self.get_split_region(fasta, chunksize)
            print('Total chunk files:' + str(len(chunklist)))
            self.mk_run_vep_script(chunklist)
        if self.opt['make_input_vcf']:
            chrom = self.opt['region'].split(':')[0]
            poslist = self.opt['region'].split(':')[1].split('-')
            spos = int(poslist[0])
            epos = int(poslist[1])
            self.save_input_vcf(fasta, self.opt['out'], chrom, spos, epos)
        if self.opt['merge_vep']:
            chunklist = self.get_split_region(fasta, chunksize)
            self.mk_shellscript_merge_vep(fasta, self.opt['out'], chunklist)
        if self.opt['check_vep_result']:
            self.check_vep_result(self.opt['vcf'], self.opt['vep_result'], 'vcf')




import re
import vcf
import argparse
import json
from .util import file_util
from .util import vcf_util
from . import _options
from .util.struct_util import get_dict_value as dv
from .model.datasource import DataSourceList
from . import external_functions

INFOIDX = vcf_util.VCF_COL.index('INFO')
FORMATIDX = vcf_util.VCF_COL.index('FORMAT')


def pars_header(line):
    d = {}
    p1 = re.compile(r'''##(?P<field>.+?)=(?P<val>.+)''')
    p2 = re.compile(r'''(?P<key>.+?)=(?P<val>[^,]+)(,|)''')
    m1 = p1.match(line)
    field = m1.group('field')
    value = m1.group('val')
    if value[0] != "<" and value[-1] != ">":
        d[field] = value
    else:
        for m2 in p2.finditer(value[1:-1]):
            key = m2.group('key')
            val = m2.group('val')
            try:
                d[field][key] = val
            except KeyError:
                d[field] = {}
                d[field][key] = val
    return d

def read_metadata_and_check_duplicate(header):
    meta = {}
    for line in header.strip().split('\n'):
        line = line.strip()
        d = pars_header(line)
        field = list(d.keys())[0]
        
        try:
            meta[field]
        except KeyError:
            if type(d[field]) == type({}):
                meta[field] = {}
            else:
                meta[field] = []

        if type(d[field]) == type({}):
            fid = d[field]['ID']
            try:
                meta[field][fid]
                raise Exception("VCF header duplicates: " + line)
            except KeyError:
                meta[field][fid] = d[field]
        else:
            if d[field] in meta[field]:
                raise Exception("VCF header duplicates: " + line)
            meta[field].append(d[field])
    return meta


class AnnotVCFValidator:
    def __init__(self, vcf_file):
        self.vcf_file = vcf_file
        self.reader = vcf.Reader(open(self.vcf_file))
        self.header = ""
        self.colnames = []
        self.mutanno_sources = {}
        self.dslist = None
        self.opt = None
        self.varkey = ""
        self.read_header()

    def set_datastructure(self, jsonfile=""):
        self.dslist = DataSourceList(jsonfile)

    def set_option(self, opt):
        if isinstance(opt, list):
            self.opt = _options.get_options()
        else:
            self.opt = opt

    def read_header(self):
        for line in file_util.gzopen(self.vcf_file):
            line = file_util.decodeb(line)
            if line[:2] == '##':
                self.header += line
            elif line[0] == '#':
                self.colnames = line.strip().split('\t')
            else:
                break

    def load_mutanno_structure(self):
        for line in self.reader._header_lines:
            if line[:len('##MUTANNO=<ID=')] == '##MUTANNO=<ID=':
                d = {}
                for f1 in line.replace('##MUTANNO=<','').replace('>','').split(','):
                    arr2 = f1.split('=')
                    d[arr2[0]] = arr2[1].replace('"','')
                self.mutanno_sources[d['ID']] = d

        for s1 in self.mutanno_sources.keys():
            try:
                self.reader.infos[s1]
            except KeyError:
                self.raise_exception("MUTANNO's " + infokey + " source doesn't have INFO field in header.")

        for infokey in self.reader.infos.keys():
            desc = self.reader.infos[infokey].desc
            if 'Format:' in desc:
                fields = desc.split("Format:'")[-1].replace("'","").split('|')
                try:
                    self.mutanno_sources[infokey]
                    self.mutanno_sources[infokey]['fields'] = fields
                except KeyError:
                    self.raise_exception("MUTANNO's " + infokey + " source doesn't have MUTANNO's version field in header.")

    def raise_exception(self, msg, print_varkey=True):
        if print_varkey:
            msg += " " + self.varkey
        msg += " in " + self.vcf_file
        raise Exception(msg)   

    def warn(self, msg, print_varkey=True):
        if print_varkey:
            msg += " " + self.varkey
        print("\tWARN: " + msg + " " + self.vcf_file)

    def check_vep_consequence_and_most_severe(self, info, col, v1=None):
        considx = col.index('Consequence')

        if 'VEP' in info.keys():
            vep_sections = []
            # print(info['VEP'])
            most_severe = 0
            for idx, attr in enumerate(info['VEP']):
                if isinstance(attr, str):
                    fields = attr.split('|')
                else:
                    fields = attr
                vep_sections.append(fields)
                if '&' in fields[considx]:
                    self.raise_exception("VEP Consequence has '&'. " + fields[considx])

                # print('col:',col)
                if 'MOST_SEVERE' in col:
                    most_severe += int(fields[col.index('MOST_SEVERE')])

            if 'MOST_SEVERE' in col and most_severe != 1:
                self.raise_exception("\tThis variant doesn't have MOST_SEVERE transcript. ")
            # for idx, fields in enumerate(vep_sections):
            #     is_most_severe = external_functions.is_most_severe_transcript(vep_sections, idx, self.mutanno_sources['VEP']['fields'])
            #     if str(is_most_severe) != str(fields[mostidx]):
            #         pass
            #         # raise Exception("VEP MOST_SEVERE is not matched." + v1.CHROM + ':' + str(v1.POS) + " " + fields[1] + " " + str(is_most_severe) + ":" + fields[mostidx])
    
    def check_genes(self, info):
        try:
            for f1 in info["GENES"]:
                if "tmpOOOOOOOOOO" in f1:
                    # fields = f1.split('|')
                    self.raise_exception("\tThe add_severe_consequence function of GENES fields doesn't work.")
        except KeyError:
            if 'VEP' in info.keys() and 'GENES' in self.dslist.sources.keys():
                self.raise_exception("\tNo GENES.")
            else:
                self.warn("No GENES.")

    def check_feature_ncbi(self, info):
        try:
            for vep in info["VEP"]:
                vepfields = vep.split('|')
                if "Feature_ncbi" in self.mutanno_sources['VEP']['fields']:
                    idx = self.mutanno_sources['VEP']['fields'].index("Feature_ncbi")
                    if vepfields[idx] == "":
                        self.warn("No Feature_ncbi")
                    else:
                        print("\tOK. Feature_ncbi=" + vepfields[idx])
        except KeyError:
            # self.raise_exception("\tNo VEP.")
            self.warn("No VEP.")


    def check_hg19(self, info):
        try:
            info["HG19"]
            for hg19 in info["HG19"]:
                if len(hg19.split('|')) != len(self.mutanno_sources['HG19']['fields']):
                    self.raise_exception("HG19 is unmatched " + hg19)
                else:
                    print('\tOK. HG19='+','.join(info["HG19"]))
        except KeyError:
            self.warn("\tNo HG19.")

    def check_samplegeno(self, info):
        try:
            # print(info["SAMPLEGENO"])
            if len(info["SAMPLEGENO"]) != len(self.opt.add_genoinfo):
                self.raise_exception("SAMPLEGENO is unmatched.")
            else:
                print('\tOK. SAMPLEGENO='+','.join(info["SAMPLEGENO"]))

        except KeyError:
            self.raise_exception("No SAMPLEGENO.")

    def validate_variant_fields_in_only_vcf(self):
        self.load_mutanno_structure()

        for r1 in self.reader:
            # varkey = r1.CHROM +":"+ str(r1.POS) + "_" + r1.REF + ">" + ",".join(r1.ALT)
            self.varkey = r1.CHROM +":"+ str(r1.POS) 

            # print("checking..", self.varkey)
            for infokey in self.mutanno_sources.keys():
                if infokey in r1.INFO.keys():
                    flag = True
                    for attr in r1.INFO[infokey]:
                        
                        fields = attr.split('|')
                        # print(">>>>",fields)
                        if len(self.mutanno_sources[infokey]['fields']) != len(fields):
                            msg = infokey + " fields is not matched. (" + '|'.join(self.mutanno_sources[infokey]['fields']) + " != " + attr + ")" 
                            self.raise_exception(msg)

                        for v1 in fields:
                            if v1 != "":
                                flag = False
                    if flag:
                        msg = infokey + " has no fields ("+infokey+"=" + attr + ")" 
                        self.raise_exception(msg)                        
                            
                    self.check_vep_consequence_and_most_severe(r1.INFO, self.mutanno_sources['VEP']['fields'], r1)

            self.check_feature_ncbi(r1.INFO)
            self.check_genes(r1.INFO)
            if self.opt is not None and self.opt.add_genoinfo:
                self.check_samplegeno(r1.INFO)
            if self.opt is not None and self.opt.add_hg19:
                self.check_hg19(r1.INFO)

    def validate(self):
        self.metadata = read_metadata_and_check_duplicate(self.header)
        self.validate_variant_fields_in_only_vcf()
        # print(self.opt)
        # print(self.mutanno_sources)

def parse_tsi_annot_header(tsi_info_header):
    h = {}
    for f1 in tsi_info_header.split(';'):
        arr = f1.split('=')
        h[arr[0]] = arr[1].split('|')
    return h

def parse_info(infofield):
    d = {}
    for f1 in infofield.split(';'):
        if '=' in f1:
            arr = f1.split('=')
            d[arr[0]] = []
            for f2 in arr[1].split(','):
                d[arr[0]].append(f2.split('|'))
        else:
            if f1 != '':
                d[f1] = True
    return d

class AnnotTSIValidator(AnnotVCFValidator):

    def __init__(self, tsi_file):
        self.vcf_file = tsi_file
        self.header = ""
        self.colnames = []
        self.mutanno_sources = {}
        self.dslist = None
        self.opt = None
        self.varkey = ""
        self.annot_header = {}
        self.read_header()

    def read_header(self):
        for line in file_util.gzopen(self.vcf_file):
            line = file_util.decodeb(line)
            if line[0] != '#':
                break
            else:
                self.colnames = line.strip().split('\t')
                self.annot_header = parse_tsi_annot_header(self.colnames[-1])
                # print(self.annot_header)


    def validate_variant_fields_in_only_tsi(self):
        for line in file_util.gzopen(self.vcf_file):
            line = file_util.decodeb(line)
            if line[0] != '#':
                arr = line.split('\t')
                infofield = parse_info(arr[-1].strip())
                for k1 in infofield.keys():
                    # print(infofield[k1])
                    for d2 in infofield[k1]:
                        if len(self.annot_header[k1]) != len(d2):
                            msg = k1 + ": the field number not matched."
                            self.raise_exception(msg)

                # print('infofield:', infofield)
                self.check_vep_consequence_and_most_severe(infofield, self.annot_header['VEP'])
        

    def validate(self):
        self.validate_variant_fields_in_only_tsi()







from .util import vcf_util
from .util.vcf_util import INFOIDX
import json


class VCFRenderer:
    def __init__(self):
        pass

    def merge_twoinfofields(self, infofield1, infofield2):
        infofield1 = vcf_util.strip_info(infofield1)
        infofield2 = vcf_util.strip_info(infofield2)
        if infofield1 == "":
            info = infofield2
        if infofield2 == "":
            info = infofield1
        if infofield1 != "" and infofield2 != "":
            info = infofield1 + ';' + infofield2
        return info

    def convert_annotdata_to_infofield(self, annotdata, available_field_list):
        info = []
        for s1 in available_field_list.keys():
            sinfo = s1 + '='
            infoattr = []
            for attr in annotdata[s1]:
                infofield = []    
                for fieldname in available_field_list[s1]:
                    if isinstance(attr[fieldname], list):
                        for i in range(len(attr[fieldname])):
                            attr[fieldname][i] = vcf_util.encode_infovalue(attr[fieldname][i])
                        infofield.append( "~".join(attr[fieldname]))
                    else:
                        attr[fieldname] = vcf_util.encode_infovalue(attr[fieldname])
                        infofield.append( attr[fieldname] )
                infoattr.append('|'.join(infofield))
            if len(infoattr) > 0:
                merged_value = ','.join(infoattr)
                if merged_value.strip().replace("|","") != "":
                    info.append(s1 + '=' + merged_value)
        return ';'.join(info)

    def render_vcfvariant(self, vcfvariant):
        if vcfvariant.is_multiallelic:
            lines = []
            for variant in vcfvariant.split_variants:
                record = variant.record
                info_record = self.render_vcfvariant_info(variant)
                record[INFOIDX] = self.merge_twoinfofields(record[INFOIDX], info_record)
                lines.append('\t'.join(record))
            return '\n'.join(lines)
        else:
            record = vcfvariant.record
            info_record = self.render_vcfvariant_info(vcfvariant)
            record[INFOIDX] = self.merge_twoinfofields(record[INFOIDX], info_record)
        return '\t'.join(record)

    def render_vcfvariant_info(self, vcfvariant):
        annotdata = vcfvariant.annotmerger.get_data()
        available_field_list = vcfvariant.annotmerger.available_field_list
        return self.convert_annotdata_to_infofield(annotdata, available_field_list)            
            

class JSONRenderer:
    def __init__(self, dslist=None, outname=None):
        self.fp = None
        self.dslist = dslist
        self.i = 0
        if outname is not None:
            self.fp = open(outname + '.json', 'w')
            self.fp.write('[')

    def __del__(self):
        if self.fp is not None:
            self.fp.write(']')
            self.fp.close()

    def save_jsonvariant(self, vcfvariant):
        if self.i > 0:
            self.fp.write(',')
        self.fp.write(self.render_jsonvariant(vcfvariant))
        self.i+=1
    
    def render_jsonvariant(self, vcfvariant):
        if vcfvariant.is_multiallelic:
            records = []
            for variant in vcfvariant.split_variants:
                record = self.render_jsonvariant(variant)
                records.append(record)
            return ','.join(records)
        else:
            annotdata = vcfvariant.annotmerger.data
            annotdata['CHROM'] = vcfvariant.chrom
            annotdata['POS'] = vcfvariant.pos
            annotdata['REF'] = vcfvariant.ref
            annotdata['ALT'] = vcfvariant.alt
            rst = json.dumps(annotdata, indent=2)
        return rst



class TSIRenderer(VCFRenderer):
    def __init__(self):
        pass

    def render_vcfvariant(self, vcfvariant, keep_empty_variant=True, print_linereturn=True):
        annotdata = vcfvariant.annotmerger.get_data()
        available_field_list = vcfvariant.annotmerger.available_field_list
        info_record = self.convert_annotdata_to_infofield(annotdata, available_field_list)
        rst = ''
        if keep_empty_variant or info_record.strip() != '':
            cont = []
            cont.append(vcfvariant.chrom)
            cont.append(str(vcfvariant.pos))
            cont.append('')
            cont.append(vcfvariant.ref)
            cont.append(vcfvariant.alt)
            cont.append(info_record)
            rst = '\t'.join(cont) 
            if print_linereturn:
                rst += '\n'
        return rst



from .. import conf
import random
import django.shortcuts


def render(request, template, data={}):
    ndata = {}
    ndata['TITLE'] = conf.TITLE
    ndata['randkey'] = random.randint(10**8, 10**9)
    ndata['errmsg'] = ""
    if "errmsg" in request.GET.keys():
        ndata['errmsg'] = request.GET['errmsg'].strip()

    ndata['sampleselect_cookie'] = request.COOKIES.get('sampleselect', {})

    for k1 in data.keys():
        ndata[k1] = data[k1]
    return django.shortcuts.render(request, template, ndata)




from ..util import file_util
from ..util import struct_util
from ..util.struct_util import get_dict_value as dv
from .. import _version
from . import datasource
import os


# (ATTRIBUTE_NAME, DEFAULT_VALUE, IS_ESSENTIAL)
DSLIST_ATTR = [
    ("name", "MUTANNO", True),
    ("version", _version.VERSION, True),
    ("version_date", _version.VERSION_DATE, True),
    ("sourcefile", "", False),
    ("sourcefile_path", "", False),
    ("single_source_mode", False, False),
    ("level", "", True),
    ("source", [], True)
]

DS_ATTR = [
    ("name", "", True),
    ("desc", "", True),
    ("version", "", True),
    ("version_date", "", True),
    ("sourcefile", "", True),
    ("function", None, True),
    ("param", None, True),
    ("subembedded", "", True),
    ("format", "tsi", True),
    ("chrompre", "", True),
    ("fieldselection", "", True),
    ("is_available", True, True),
    ("datatype", "variant", True),
    ("ref_column_index", 3, True),
    ("alt_column_index", 4, True),
    ("fields", [], True)
]

DSFIELD_ATTR = [
    ("name", "", True),
    ("name2", "", True),
    ("type", "string", True, ["string", "number", "integer", "boolean"]),
    ("desc", "", False),
    ("is_list", False, True),
    ("delimiter", "~", True),
    ("function", None, True),
    ("param", None, True),
    ("default", None, True),
    ("subembedded_key", False, True),
    ("keep_equal_str", [], True),
    ("is_available", True, True),
]


class DataSourceListStructure:
    def __init__(self, dsjsonfile=""):
        if isinstance(dsjsonfile, str) and file_util.is_exist(dsjsonfile):
            self.source_list = []
            self.available_source_list = []
            self.sources = {}
            self.dsjsonfile = dsjsonfile
            self.sourcefile2_list = []
            self.load_dsfile()
            self.update_sourcefile2()
            if len(self.sourcefile2_list) == 0:
                self.single_source_mode = True


    def load_dsfile(self):
        ds = file_util.load_json(self.dsjsonfile)
        for attr in DSLIST_ATTR:
            if attr[0] == "source":
                for source_ds in ds["source"]:
                    dss = datasource.DataSource(source_ds)
                    dss.update_sourcefile2(self.sourcefile_path)
                    if dss.sourcefile2 != '':
                        self.sourcefile2_list.append(dss.sourcefile2)

                    if dss.is_available:
                        self.source_list.append(dss)
                        if dss.is_available:
                            self.available_source_list.append(dss)
                        self.sources[dss.name] = dss
            else:
                self.__dict__[attr[0]] = dv(ds, attr[0], attr[1])

    def update_sourcefile2(self):
        if self.sourcefile_path != "":
            self.sourcefile2 = os.path.join(self.sourcefile_path, self.sourcefile)
        else:
            self.sourcefile2 = self.sourcefile

class DataSourceStructure:
    def __init__(self, dsjson=None):
        self.field_list = []
        self.available_field_list = []
        self.fields = {}
        self.fields2 = {}
        self.field_name2name1 = {}
        if isinstance(dsjson, dict):
            for attr in DS_ATTR:
                if attr[0] == "fields":
                    for field_ds in dsjson["fields"]:
                        dsfs = DataSourceFieldStructure(field_ds)
                        self.field_list.append(dsfs)
                        self.fields[dsfs.name] = dsfs
                        self.field_name2name1[dsfs.name2] = dsfs.name

                        if dsfs.is_available:
                            self.available_field_list.append(dsfs)
                else:
                    self.__dict__[attr[0]] = dv(dsjson, attr[0], attr[1])

    def __str__(self):
        return self.name
    
    def update_sourcefile2(self, sourcefile_path):
        # print(">DataSourceStructure.update_sourcefile2()", self.name)
        if sourcefile_path != "":
            self.sourcefile2 = os.path.join(sourcefile_path, self.sourcefile)
        else:
            self.sourcefile2 = self.sourcefile
        # print(self.sourcefile2)


class DataSourceFieldStructure:
    def __init__(self, dsjson=None):
        if isinstance(dsjson, dict):
            for attr in DSFIELD_ATTR:
                self.__dict__[attr[0]] = dv(dsjson, attr[0], attr[1])
            self.set_name2()

    def __str__(self):
        return self.name
    
    def set_name2(self):
        if self.name2 == "":
            self.name2 = self.name


    def convert_type_from_string(self, strvalue):
        if self.type == "number":
            if strvalue == "":
                rst = self.default
            else:
                rst = float(strvalue)
        elif self.type == "integer":
            if strvalue == "":
                rst = self.default
            else:
                rst = int(strvalue)
        elif self.type == "boolean":
            if strvalue == "1":
                rst = True
            elif strvalue == "0":
                rst = False
            else:
                rst = None
        else:
            if strvalue == "":
                rst = self.default
            else:
                rst = strvalue
        return rst
        


from .util import file_util
from .util import vep_util
from .util import struct_util


def conv_consequence_delimiter(consequence):
    rst = consequence
    if isinstance(consequence, str):
        rst = consequence.replace('&', '~')
    elif isinstance(consequence, list):
        rst = []
        for c1 in consequence:
            rst.append(c1.replace('&', '~'))
    return rst


def get_score_from_vep_pathogenicity(predscore):
    rst = predscore
    if isinstance(predscore, str):
        arr = predscore.replace(')', '').split('(')
        rst = arr[1].strip()
    return rst


def get_pred_from_vep_pathogenicity(predscore):
    rst = predscore
    if isinstance(predscore, str):
        arr = predscore.replace(')', '').split('(')
        rst = arr[0].strip()
    return rst


def add_vep_most_severe(sections):
    rst = '0'
    # print (">external_functions.is_most_severe_transcript():", rst)
    most_severe = {}
    most_severe['corder'] = 9999
    target_gene = ""
    for sidx, section in enumerate(sections):
        d = {}
        # print("\tsection:",section)

        for consequence in section['Consequence']:
            if (
                (most_severe['corder'] > vep_util.VEP_CONSEQUENCE_ORDER[consequence])
                or (
                    most_severe['corder'] == vep_util.VEP_CONSEQUENCE_ORDER[consequence]
                    and section['CANONICAL'] is not None and section['CANONICAL'] == "1"
                    )
                ):
                most_severe['sidx'] = section['Gene']
                most_severe['sidx'] = sidx
                most_severe['canonical'] = section['CANONICAL']
                most_severe['corder'] = vep_util.VEP_CONSEQUENCE_ORDER[consequence]
            # elif most_severe['corder'] == vep_util.VEP_CONSEQUENCE_ORDER[consequence]:
            #     if section['CANONICAL'] is not None and section['CANONICAL'] == "1":
            # print("\t\tconsequence:",consequence)
        # print("\tmost_severe:",most_severe)

    rst = []
    ms = ''
    for sidx, section in enumerate(sections):
        if sidx == most_severe['sidx']:
            # print('====>', sections[sidx])
            sections[sidx]['MOST_SEVERE'] = '1'
            # ms = '1'
        else:
            sections[sidx]['MOST_SEVERE'] = '0'
            # ms = '0'
    # print(sections)
    # return ms
    return sections

def vep_select_biotype_add_most_severe(sections, select_biotype):
    sections = vep_select_biotype(sections, select_biotype)
    sections = add_vep_most_severe(sections)
    return sections
    

def vep_select_biotype(sections, select_biotype):
    # print (">external_functions.vep_select_biotype():", select_biotype)
    # print(sections)
    selected = []
    for sidx, section in enumerate(sections):
        # print("section['BIOTYPE']:",section['BIOTYPE'])
        if section['BIOTYPE'] in select_biotype:
            selected.append(section)
    return selected

def trim_DIP_ID(v1):
    # ex) DIP-39616N;
    rst = v1
    if isinstance(v1, str):
        rst = v1.split('-')[-1].replace('N', '')
    return rst


def remove_pdb_subversion(id_list):
    rst = id_list
    if isinstance(id_list, str):
        pidmap = {}
        for pid in id_list.split('|'):
            pidmap[pid.split(':')[0]] = 1
        pidlist = list(pidmap.keys())
        rst='|'.join(pidlist)
    return 


def convert_rmsk_strand(v1):
    rst = '0'
    if isinstance(v1, str):
        if v1 == '+':
            rst = '1'
    return rst


def convert_NS2blank(v1):
    rst = v1
    if isinstance(v1, str):
        if v1 == 'NS':
            rst = ''
    return rst


def convert_vep_strand(v1):
    rst = v1
    if isinstance(v1, str):
        rst = v1.replace('-1', '0')
    return rst


def convert_high_inf_pos(v1):
    rst = v1
    if isinstance(v1, str):
        v1.replace('Y', '1').replace('N', '0')
    return rst


def convert_canonical2boolean(canonical):
    rst = '0'
    if isinstance(canonical, str):
        if canonical == 'YES':
            rst = '1'
    return rst


def convert_uniprot_transmem(desc_value):
    rst = desc_value
    if isinstance(desc_value, str):
        arr = []
        for v1 in desc_value.split(';'):
            arr.append(v1.strip())
        rst = ";".join(arr)
    return rst


def add_genes_severe_consequence(annotdata, vcfinfo):
    # print(">external_function.add_severe_consequence()")
    # print('\t',annotdata['VEP'])

    vep_sections = []
    if annotdata is not None and 'VEP' in annotdata.keys():
        vep_sections = annotdata['VEP']
    elif vcfinfo is not None and 'VEP' in vcfinfo.keys():
        vep_sections = vcfinfo['VEP']

    for attr in vep_sections:
        if attr['MOST_SEVERE'] == '1':
            ensg = attr['Gene']
            most_severe_transcript = attr['Feature']
            most_severe_consequence = attr['Consequence']
            break
    if len(vep_sections) == 0:
        rst_sections = []
    else:
        rst_sections =[{'ensg':ensg, 'most_severe_transcript':most_severe_transcript, 'most_severe_consequence':most_severe_consequence}]
    return rst_sections


# entrezmap = {}
# refseqmap = {}
# def load_entrez_refseq_id():
#     path = "/home/mk446/bio/mutanno/DATASOURCE/ENSEMBL/hg38/"
#     entrezmap = {}
#     for line in file_util.gzopen(path + 'Homo_sapiens.GRCh38.98.entrez.tsv.gz'):
#         line = line.decode('UTF-8')
#         arr = line.split('\t')
#         arr[-1] = arr[-1].strip()
#         entrezmap[arr[0].strip()] = arr[3].strip()
#     refseqmap = {}
#     for line in file_util.gzopen(path + 'Homo_sapiens.GRCh38.98.refseq.sorted.tsv.gz'):
#         line = line.decode('UTF-8')
#         arr = line.split('\t')
#         arr[-1] = arr[-1].strip()
#         enst_id = arr[1].strip()
#         refseqmap[enst_id] = arr[3].strip()
#     return entrezmap, refseqmap


# def get_entrez_id(gene):
#     global entrezmap, refseqmap
#     if len(entrezmap.keys()) == 0:
#         entrezmap, refseqmap = load_entrez_refseq_id()
#     try:
#         eid = entrezmap[gene]
#     except KeyError:
#         eid = ''
#     return eid


# def get_refseq_id(transcriptid):
#     global entrezmap, refseqmap
#     if len(refseqmap.keys()) == 0:
#         entrezmap, refseqmap = load_entrez_refseq_id()
#     try:
#         tid = refseqmap[transcriptid]
#     except KeyError:
#         tid = ''
#     return tid



from os.path import isfile
from ..util import file_util
from ..util import struct_util
from . import datastructure
from .. import external_functions
from .datastructure import DataSourceStructure, DataSourceListStructure
import tabix


def run_external_function(function, params, paramvalues):
    # print("paramvalues", paramvalues['mutanno_value_sections'])

    exec_str = "external_functions." + function
    exec_str += '('
    paramstr_list = []
    for param in params:
        # print('param:', param)
        if param in paramvalues.keys():
            paramstr_list.append('paramvalues["' + param + '"]')
            # print('paramstr_list:', paramstr_list)
        else:
            paramstr_list.append(param)
    exec_str += ','.join(paramstr_list)
    exec_str += ')'
    # print('@@@@',exec_str)
    # print(paramvalues)
    rstvalue = eval(exec_str)
    return rstvalue


class VariantAnnotMerger:
    def __init__(self):
        self.data = {}
        self.annot = {}
        self.annot_list = []
        self.available_field_list = {}
        self.flag_set_data = False

    def add_variant_annot(self, variant_annot):
        self.annot[variant_annot.source.name] = variant_annot
        self.annot_list.append(variant_annot)

    def get_data(self):
        if not self.flag_set_data:
            self.set_data()
        return self.data

    def set_data(self):
        for annot in self.annot_list:
            annotdata = annot.get_data(self.data)
            sname = annot.source.name
            try:
                self.data[sname]
            except KeyError:
                self.data[sname] = []
            self.data[sname].extend(annotdata)

            try:
                self.available_field_list[sname].extend(annot.available_field_list)
            except KeyError:
                self.available_field_list[sname] = annot.available_field_list
        self.flag_set_data = True


class VariantAnnot:
    def __init__(self, records, header_record, datasource=None, return_all=False, variant=None) :
        self.data = []
        self.rawdata = []
        self.records = records
        self.header_record = header_record
        self.header = []
        self.source = datasource # DataSource
        self.return_all = return_all
        self.available_field_list = []
        self.flag_set_data = False
        self.othersource_annotdata = {}
        self.variant = variant
        self.vcf_info = None
        if self.variant is not None:
            self.vcf_info = self.variant.get_vcf_info()
        self.set_rawdata()

    def get_data(self, othersource_annotdata):
        self.othersource_annotdata = othersource_annotdata
        if not self.flag_set_data:
            self.set_data()
        return self.data

    def set_rawdata(self):
        # print("### set_rawdata():" + self.source.name)
        if len(self.rawdata) == 0:
            if self.source.format == "tsi":
                self.set_rawdata_from_tsi()
            if self.source.format == "bed":
                if self.return_all or self.source.is_available:
                    self.set_rawdata_from_bed()
            if self.source.format == "tsv":
                if self.return_all or self.source.is_available:
                    self.set_rawdata_from_tsv()
            if self.source.format == "function":
                if self.return_all or self.source.is_available:
                    self.set_rawdata_from_function()

    def set_data(self):
        self.set_data_from_rawdata()
        self.flag_set_data = True

    def set_data_from_rawdata(self):
        # print("### set_data_from_rawdata(): " + self.source.name)
        if self.source.is_available:
            for f1 in self.source.available_field_list:
                self.available_field_list.append(f1.name)
            # print(self.rawdata)
            # print("\t===(1)self.rawdata:", self.rawdata)
            self.rawdata = self.apply_external_function(self.rawdata)
            # print("\t===(2)self.rawdata:", self.rawdata)
            for attr in self.rawdata:
                d = {}
                for f1 in self.source.available_field_list:
                    # if self.source.fields[f1.name].function is None:
                    try:
                        d[f1.name] = attr[f1.name]
                        # print(">>>", f1.name, attr[f1.name])
                    except KeyError:
                        if f1.default != None:
                            d[f1.name] = f1.default
                        else:  
                            d[f1.name] = ""
                self.data.append(d)
            if len(self.data) == 0:
                d = {}
                for f1 in self.source.available_field_list:
                    if f1.default != None:
                        d[f1.name] = f1.default
                if len(d.keys()) > 0:
                    self.data = [d]
            # print("\tself.data:", self.data)


    def apply_external_function(self, sections):
        # print(">>>sections",sections)
        for section_idx, section in enumerate(sections):
            # print(">>>>section", section)

            if self.source.function is not None:
                function = self.source.function
                params = self.source.param.split(',')
                paramvalues = {
                    "mutanno_value_sections":sections,
                    "mutanno_value_section_idx":section_idx,
                    "mutanno_value_columnheader":self.header,
                    "vcf_info_value":self.vcf_info,
                    "mutanno_value_data":self.othersource_annotdata
                }
                for k1 in section.keys():
                    paramvalues[k1] = section[k1]
                sections = run_external_function(function, params, paramvalues)

            for field in self.source.available_field_list:
                try:
                    if self.source.fields[field.name].function is not None:

                        field = self.source.fields[field.name]
                        function = field.function
                        # print("########", function)
                        params = field.param.split(',')
                        paramvalues = {
                            "mutanno_value_sections":sections,
                            "mutanno_value_section_idx":section_idx,
                            "mutanno_value_columnheader":self.header,
                            "vcf_info_value":self.vcf_info,
                            "mutanno_value_data":self.othersource_annotdata
                        }
                        for k1 in section.keys():
                            paramvalues[k1] = section[k1]
                        section[field.name] = run_external_function(function, params, paramvalues)
                except KeyError:
                    pass
        return sections
    
    def set_rawdata_from_tsi(self):
        """
        The raw data is unselected values from data source.
        """
        if '=' in self.header_record[-1]:
            for s1 in self.header_record[-1].split(';'):
                arr = s1.split('=')
                source_name = arr[0]
                if source_name == self.source.name:
                    values = arr[1].split('|')
                    self.header = arr[1].split('|')
        else:
            self.header = self.header_record[-1].split('|')

        self.header = self.source.conv_fieldname1(self.header)

        for rec in self.records:
            for s1 in rec[-1].split(';'):
                if '=' in s1:
                    arr = s1.split('=')
                    source_name = arr[0]
                    r1s = arr[1].split(',')
                    source_is_available = True
                else:
                    source_name = self.source.name
                    source_is_available = self.source.is_available
                    r1s = s1.split(',')

                if source_name == self.source.name:
                    
                    if self.return_all or source_is_available:
                        for r1 in r1s:
                            values = r1.split('|')

                            d = self.set_init_annot_section_key(rec)
                            for i, fieldname in enumerate(self.header):
                                d[fieldname] = self.convert_field_type(fieldname, values[i])
                                
                            if len(d.keys()) > 0:
                                self.rawdata.append(d)


    def set_init_annot_section_key(self, rec):
        d = {}
        d['CHROM'] = rec[0]
        d['POS'] = int(rec[1])
        d['REF'] = rec[self.source.ref_column_index]
        d['ALT'] = rec[self.source.alt_column_index]
        d['VARKEY'] = d['CHROM'] + ':' + str(d['POS']) + '_' + d['REF'] + '>' + d['ALT']
        return d

    def check_field_availability(self, fieldname):
        flag = False
        try:
            self.source.fields[fieldname]
            if self.return_all or self.source.fields[fieldname].is_available:
                flag = True
        except KeyError:
            flag = self.return_all
        return flag

    def convert_field_type(self, fieldname, value):
        try:
            fieldstructure = self.source.fields[fieldname]
            if fieldstructure.is_list:
                if value == "":
                    rst = None
                else:
                    rst = []
                    for v1 in value.split(fieldstructure.delimiter):
                        rst.append(fieldstructure.convert_type_from_string(v1))
            else:
                rst = fieldstructure.convert_type_from_string(value)
        except KeyError:
            if value == "":
                rst = None
            else:
                rst = value
        return rst

    def set_rawdata_from_bed(self):
        # print("???1", self.source.name, self.records)
        self.header = []
        for rec in self.records:
            d = self.set_init_annot_section_key(rec)
            for i, fieldname in enumerate(self.header_record):
                # if self.check_field_availability(fieldname):
                d[fieldname] = self.convert_field_type(fieldname, rec[i])
                self.header.append(fieldname)
            self.rawdata.append(d)

    def set_rawdata_from_tsv(self):
        self.header = []
        for rec in self.records:
            d = self.set_init_annot_section_key(rec)
            for i, fieldname in enumerate(self.header_record):
                # if self.check_field_availability(fieldname):
                d[self.header_record[i]] = self.convert_field_type(fieldname, rec[i])
                self.header.append(fieldname)
            self.rawdata.append(d)

    def set_rawdata_from_function(self):
        # print(">VariantAnnot.set_rawdata_from_function() : " + self.source.name)
        d = {}
        for field in self.source.field_list:
            d[field.name] = "tmpOOOOOOOOOO"
            self.header.append(field.name)
        self.rawdata.append(d)

    
class DataSource(DataSourceStructure):
    tchrom = ""
    source_name = ""
    header = ""
    tabixpointer = None
    
    def set_header(self, sourcefile):
        # print(">DataSource.set_header():", self.name)
        if self.header == "":
            for line in file_util.gzopen(sourcefile):
                line = file_util.decodeb(line)
                if line[0] == "#":
                    if line[:2] != "##":
                        self.header = file_util.line2arr(line[1:])
                        break

    def conv_fieldname1(self, tmp_name_list):
        if isinstance(tmp_name_list, str):
            tmp_name_list = [tmp_name_list]
        rstlist = []
        for tname in tmp_name_list:
            try:
                rstlist.append(self.field_name2name1[tname])
            except KeyError:
                rstlist.append(tname)
        return rstlist

    def tabix_load(self, chrom=""):
        # print(">DataSource.tabix_load():", self.name)
        sourcefile = self.sourcefile2.replace('#CHROM#', chrom)
        # print(">>", sourcefile, self.name)
        if file_util.is_exist(sourcefile) and isfile(sourcefile):
            # print("loading..", sourcefile)
            self.tabixpointer = tabix.open(sourcefile)
            self.tchrom = chrom
            self.set_header(sourcefile)
        else:
            if self.sourcefile != "":
                raise Exception("Source file is not exist. : " + sourcefile )

    def get_variant_annot(self, chrom, pos, ref="", alt="", epos=-9, return_raw=True, records = [], header_record = [], variant=None):
        # print(self.sourcefile2)
        if self.tchrom != chrom and (self.tabixpointer == None  or "#CHROM#" in self.sourcefile2):
            self.tabix_load(chrom)

        if len(header_record) > 0:
            recs = records
            header = header_record
        else:
            if epos > pos:
                chrompos = chrom + ':' + str(pos) + '-' + str(epos)
            else:
                chrompos = chrom + ':' + str(pos) + '-' + str(pos)
            recs = []
            try:
                for rec in self.tabixpointer.querys(chrompos):
                    if (epos < 0 and rec[self.ref_column_index] == ref and rec[self.alt_column_index] == alt
                        or (epos > pos)
                    ):
                        recs.append(rec)
            except AttributeError:
                pass
            except tabix.TabixError:
                pass
            header = self.header

        return VariantAnnot(recs, header, datasource=self, variant=variant)

    def get_variantkeys(self, chrom, spos, epos):
        if self.tchrom != chrom and (self.tabixpointer == None  or "#CHROM#" in self.sourcefile2):
            self.tabix_load(chrom)
        chrompos = chrom + ':' + str(spos) + '-' + str(epos)
        vkeys = {}
        if self.tabixpointer is not None:
            for rec in self.tabixpointer.querys(chrompos):
                vkey = rec[0] + "_" + rec[1] + "_" + "_" + rec[self.ref_column_index] + "_" + rec[self.alt_column_index]
                pos = int(rec[1])
                try:
                    vkeys[pos]
                except KeyError:
                    vkeys[pos] = []
                if vkey not in vkeys[pos]:
                    vkeys[pos].append(vkey)
            
        return vkeys



class DataSourceList(DataSourceListStructure):

    def get_variant_annot(self, chrom, pos, ref="", alt="", epos=-9, variant=None):
        chrompos = chrom + ':' + str(pos) + '-' + str(pos)
        annotmerger = VariantAnnotMerger()

        if self.sourcefile != "": 
            singlesource_ds = {
                "name": "MUTANNO_SINGLESOURCE",
                "sourcefile": self.sourcefile,
                "format": "tsi",
                "is_available": True,
                "fields": []
            }
            self.single_source = DataSource(singlesource_ds)
            self.single_source.sourcefile2 = self.sourcefile2
            single_source_annot = self.single_source.get_variant_annot(chrom, pos, ref, alt, return_raw=True)
            # annotmerger.add_variant_annot(single_source_annot)

        for s1 in self.source_list:
            if self.single_source_mode:
                annot = s1.get_variant_annot(chrom, pos, ref=ref, alt=alt, epos=epos, records=single_source_annot.records, header_record=single_source_annot.header_record, variant=variant )
            else:
                annot = s1.get_variant_annot(chrom, pos, ref=ref, alt=alt, epos=epos, variant=variant)
            annotmerger.add_variant_annot(annot)

        return annotmerger

    def get_variantkeys(self, chrom, spos, epos):
        merged_vkeys = {}
        for s1 in self.source_list:
            if s1.format in ["tsi", "tsv"]:
                vkeys = s1.get_variantkeys(chrom, spos, epos)
                for pos in vkeys.keys():
                    try:
                        merged_vkeys[pos]
                    except KeyError:
                        merged_vkeys[pos] = []

                    for vkey in vkeys[pos]:
                        if vkey not in merged_vkeys[pos]:
                            merged_vkeys[pos].append(vkey)
        return merged_vkeys


# class MutannoDataParser:
#     def __init__(self):
#         pass

#     def set_header(self, line):
#         self.header = file_util.line2arr(line)
#         self.header[0] = self.header[0].replace('#','')
#         self.annotheader = {}
#         for f1 in self.header[-1].split(';'):
#             arrf1 = f1.split('=')
#             self.annotheader[arrf1[0]] = arrf1[1].split('|')

#     def parse(self, line):
#         arr = file_util.line2arr(line)
#         dat = {}
#         for i in range(len(arr)-1):
#             dat[self.header[i]] = arr[i]
#         dat['POS'] = int(dat['POS'])
#         for f1 in arr[-1].split(';'):
#             arrf1 = f1.split('=')
#             d = []

#             for f11 in arrf1[1].split(','):
#                 arrf2 = f11.split('|')
#                 d2 = {}
#                 for i in range(len(self.annotheader[arrf1[0]])):
#                     d2[self.annotheader[arrf1[0]][i]] = arrf2[i]
#                 d.append(d2)
#             dat[arrf1[0]] = d
#         return dat

# class MutannoDataSequentialReader:
#     def __init__(self, source_file):
#         self.source_file = source_file
#         self.fp = file_util.gzopen(source_file)
#         self.mtparser = MutannoDataParser()
#         self.load_header()

#     def __iter__(self):
#         return self

#     def load_header(self):
#         line = file_util.decodeb(self.fp.readline())
#         if line[0] == '#':
#             self.mtparser.set_header(line)

#     def __next__(self):
#         line = file_util.decodeb(self.fp.readline())
#         if line.strip() == '':
#             raise StopIteration
#         elif line[0] == '#':
#             self.mtparser.set_header(line)
#         else:
#             mtdata = self.mtparser.parse(line)
#             return (mtdata)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import argparse
from . import _version

PROG = "mutanno"

DEFAULT_OPT = {'annot': {}, 'makedata': {}, 'convert': {}}
DEFAULT_OPT['annot']['vcf'] = ""
DEFAULT_OPT['annot']['out'] = "mutanno"
DEFAULT_OPT['annot']['outtype'] = ['vcf']
DEFAULT_OPT['annot']['ds'] = ""
DEFAULT_OPT['annot']['sourcefile'] = ""
DEFAULT_OPT['annot']['add_genoinfo'] = False
DEFAULT_OPT['annot']['hgvs'] = False
DEFAULT_OPT['annot']['variant_class'] = False
DEFAULT_OPT['annot']['hg19'] = False
DEFAULT_OPT['annot']['genetable'] = False
DEFAULT_OPT['annot']['split_multi_allelic_variant'] = False
DEFAULT_OPT['annot']['single_source_mode'] = False
DEFAULT_OPT['annot']['load_source_in_memory'] = False
DEFAULT_OPT['annot']['sparse'] = False
DEFAULT_OPT['annot']['silence'] = False
DEFAULT_OPT['annot']['debug'] = False


def get_options():
    global DEFAULT_OPT

    parser = argparse.ArgumentParser(
        usage='%(prog)s <sub-command> [options]', description='%(prog)s ver' + _version.VERSION + " (" +
        _version.VERSION_DATE + ")" + ': python tool for variant annotation')
    parser.add_argument('-v', '--version', action='version',
                        version="%(prog)s ver" + _version.VERSION + " (" + _version.VERSION_DATE + ")")
    subparsers = parser.add_subparsers(
        title="sub-commands", dest="subcommand", metavar='', prog=PROG)

    do = DEFAULT_OPT['annot']
    p1 = subparsers.add_parser('annot', help='annotation', description='annotation')
    p1.add_argument('-vcf', dest='vcf', default=do['vcf'], help='VCF file')
    p1.add_argument('-out', dest='out', default=do['out'], help='title of output file')
    p1.add_argument('-outtype', dest='outtype', default=do['outtype'],
                    choices=['vcf', 'json'], help='output file type', nargs='*')
    p1.add_argument('-ds', dest='ds', default=do['ds'], help='data structure json file')
    p1.add_argument('-sourcefile', dest='sourcefile', default=do['sourcefile'], help='data source file')
    p1.add_argument('-genoinfo', dest='add_genoinfo',
                    default=do['add_genoinfo'], help='add genotype info. in INFO field', nargs="*")
    p1.add_argument('-hgvs', dest='add_hgvs',
                    action="store_true", default=do['hgvs'], help='add hgvs')
    p1.add_argument('-variant_class', dest='add_variant_class',
                    action="store_true", default=do['variant_class'], help='add variant class')
    p1.add_argument('-hg19', dest='add_hg19',
                    action="store_true", default=do['hg19'], help='add hg19 coordinates')
    p1.add_argument('-chain', dest='chain',
                    default="", help='chain file for liftover of hg19 coordinates')
    p1.add_argument('-genetable', dest='add_genetable',
                    action="store_true", default=do['genetable'], help='add gene table')
    p1.add_argument('-blocksize', dest='blocksize', default='', help='block size')
    p1.add_argument('-split_multi_allelic_variant', dest='split_multi_allelic_variant',
                    action="store_true", default=do['split_multi_allelic_variant'], help='split multi-allelic variants')
    p1.add_argument('-clean_tag', dest='clean_tag_list',
                    default=[], help='remove previous annotation information', nargs='*')
    p1.add_argument('-single_source_mode', dest='single_source_mode',
                    action="store_true", default=do['single_source_mode'], help='single source mode')
    p1.add_argument('-load_source_in_memory', dest='load_source_in_memory',
                    action="store_true", default=do['load_source_in_memory'], help='loading data source in memory')
    p1.add_argument('-sparse', dest='sparse',
                    action="store_true", default=do['sparse'], help='for sparse variant position')
    p1.add_argument('-log', dest='logfile', default='', help='log file')
    p1.add_argument('-silence', dest='silence', action="store_true",
                    default=do['silence'], help='do not print any log.')
    p1.add_argument('-debug', dest='debug', action="store_true",
                    default=do['debug'], help='turn on the debugging mode')

    do = DEFAULT_OPT['makedata']
    p1 = subparsers.add_parser('makedata', help='make a single data source file',
                               description='make a single data source file')
    p1.add_argument('-out', dest='out', default='', help='title of output file')
    p1.add_argument('-outtype', dest='outtype', default='json',
                    choices=['tsv', 'json'], help='output file type (default: tsv)')
    p1.add_argument('-ds', dest='ds', default='', help='datasource json file')
    p1.add_argument('-region', dest='region', default='',
                    help='target region: (ex -region chr1:12345678-22345678 )')
    p1.add_argument('-region_vcf', dest='region_vcf', default='',
                    help='target regions using vcf file')
    p1.add_argument('-blocksize', dest='blocksize', default='', help='block size')
    p1.add_argument('-vartype', dest='vartype', default='all',
                    choices=['SNV', 'GENE', 'GENE_MAIN_CHROM', 'CODING_GENE', 'CODING_GENE_MAIN_CHROM'],
                    help='variant type')
    p1.add_argument('-debug', dest='debug', action="store_true",
                    default=False, help='turn on the debugging mode')
    p1.add_argument('-check', action="store_true", dest='check', default=False, help='check output file')
    p1.add_argument('-log', dest='logfile', default='', help='log file')
    p1.add_argument('-silence', dest='silence', action="store_true", default=False, help='do not print any log.')

    do = DEFAULT_OPT['convert']
    p1 = subparsers.add_parser('convert', help='convert', description='convert')
    p1.add_argument('-vcf2tsv', dest='vcf2tsv', action="store_true", default=False, help='convert vcf to tsv format')
    p1.add_argument('-vcf2html', dest='vcf2html', action="store_true", default=False, help='convert vcf to html format')
    p1.add_argument('-vcf2json', dest='vcf2json', action="store_true", default=False, help='convert vcf to json format')
    p1.add_argument('-tsi2json', dest='tsi2json', action="store_true", default=False, help='convert vcf to json format')
    # p1.add_argument('-vep2tab', dest='vep2tab', action="store_true", default=False, help='convert vep to tsv format')
    p1.add_argument('-in', dest='infile', default='', help='input file')
    p1.add_argument('-out', dest='outfile', default='', help='title of output file')
    p1.add_argument('-ds', dest='ds', default='', help='datastructure json file')
    p1.add_argument('-region', dest='region', default='',
                    help='target region: (ex -region chr1:12345678-22345678 )')
    p1.add_argument('-chromsplit', dest='chromsplit', action="store_true",
                    default=False, help='save separate files by chromosomes')
    p1.add_argument('-silence', dest='silence', action="store_true", default=False, help='do not print any log.')
    # p1.add_argument('-debug', dest='debug', action="store_true", default=False, help='turn on the debugging mode')

    # p1 = subparsers.add_parser('make_datasource', help='convert', description='convert')
    # p1.add_argument('-ds', dest='ds', default='datastructure.json', help='datasource json file')
    # p1.add_argument('-variant', dest='variant', default='datastructure.json', help='datasource json file')

    p1 = subparsers.add_parser('precal', help='pre-calculate', description='pre-calculate')
    p1.add_argument('-make_input_vcf', dest='make_input_vcf', action="store_true",
                    default=False, help='make input VCF file')
    p1.add_argument('-merge_vep', dest='merge_vep',
                    action="store_true", default=False, help='check and merge VEP result')
    p1.add_argument('-check_vep_result', dest='check_vep_result',
                    action="store_true", default=False, help='check VEP result')
    p1.add_argument('-run_vep', dest='run_vep', action="store_true", default=False, help='run vep')
    p1.add_argument('-out', dest='out', default='', help='title of output file')
    p1.add_argument('-fasta', dest='fasta', default='', help='reference')
    p1.add_argument('-vep', dest='vep', default='', help='vep path')
    p1.add_argument('-vepcache', dest='vepcache', default='', help='vep cache directory path')
    p1.add_argument('-cache_version', dest='cache_version', default='98', help='vep cache version')
    p1.add_argument('-region', dest='region', default='', help='region (chr1:123456-789012)')
    p1.add_argument('-vcf', dest='vcf', default="", help='VCF file')
    p1.add_argument('-vep_result', dest='vep_result', default='', help='vep result file')

    p1 = subparsers.add_parser('validate', help='validate data format',
                               description='validate data format')
    p1.add_argument('-vcf', dest='vcf', default='', help='VCF file')
    # p1.add_argument('-out', dest='out', default='', help='title of output file')
    # p1.add_argument('-silence', dest='silence', action="store_true", default=False, help='don\'t print any log.')
    # p1.add_argument('-debug', dest='debug', action="store_true", default=False, help='turn on the debugging mode')

    p1 = subparsers.add_parser('preprocess', help='quality metrics for VCF',
                               description='quality metrics for VCF')
    p1.add_argument('-infile', dest='infile', default='', help='title of input file')
    p1.add_argument('-out', dest='out', default='', help='title of output file')
    p1.add_argument('-ds', dest='ds', default='', help='datasource json file')
    p1.add_argument('-make_dbnsfp_transcript', dest='make_dbnsfp_transcript', default=False, action="store_true",
                    help='make dbNSFP transcript file')

    p1 = subparsers.add_parser('web', help='web view',
                               description='web view')
    p1.add_argument('-ds', dest='ds', default='', help='datasource json file')
    p1.add_argument('-port', dest='port', default='8080', help='port')
    p1.add_argument('-log', dest='logfile', default='', help='log file')
    p1.add_argument('-silence', dest='silence', action="store_true", default=False, help='do not print any log.')
    p1.add_argument('-debug', dest='debug', action="store_true", default=False, help='turn on the debugging mode')

    if len(sys.argv) == 1 or (len(sys.argv) == 2 and sys.argv[1][0] != '-'):
        sys.argv.append('-h')
    opt = parser.parse_args()
    # opt = vars(parser.parse_args())
    return opt


def get_opt_object_from_dict(optdict={}):
    opt = Option()
    opt.convert_from_dict(optdict)
    return opt

class Option:
    def __init__(self):
        pass

    def convert_from_dict(self, optdict={}):
        for k1, v1 in optdict.items():
            self.__dict__[k1] = v1



VERSION = "0.4.1"
VERSION_DATE = "2020.06.08"

# MUTANNO version
# 0.4.1 (2020.06.08) :
#   - turn to be 'single_source_mode = True' when data structure file has only single source file. 
#   - allocate all sample names when users use -genoinfo without indicating sample names. 
# 0.4.0 (2020.06.05) :
#   - add variant type
#   - support multiple datasources for vcf annotation
#   - remove block-size option
#   - make `MUTANNO` source name for 'samplevariantkey', 'hgvsg', and 'variant_class'
#   - add '-variant_class' option
#   - refactoring
#	- add multi-source mode
#	- refactoring makedata mode
# 0.3.19 (2020.05.25) :
#   - split pred and score in pathogenicity score from VEP (SIFT and Polyphen)
# 0.3.18 (2020.05.13) :
#   - add chain file option for pyliftover
# 0.3.17 (2020.05.06) :
#   - add is_most_severe_transcript in external function.py
#   - add keep_equal_str filter in data structure file
#   - update -geno_info option to get sample ID list (ex. -geno_info HG002 HG003 HG004 )
#   - update hgvs chromosome.
# 0.3.16 (2020.04.27) :
#   - add GENE_MAIN_CHROM, CODING_GENE_MAIN_CHROM in -vartype option
# 0.3.15 (2020.03.26) :
#   - add 'web' option
# 0.3.14 (2020.03.21) :
#   - implement pypi installation
#   - apply pytest
# 0.3.13 (2020.03.10) :
#   - add 'add_genoinfo' option
#   - add 'split_multi_allelic_variant' option
#   - add 'clean_tag' option
# 0.3.12 (2020.03.05) :
#   - add 'default' value
# 0.3.11 (2020.02.28) :
#   - add SAMPLEGENO tag
#   - minor fix in header
# 0.3.10 (2020.02.26) :
#   - minor update in version tag
# 0.3.9 (2020.02.26) :
#   - remove unannotated variant (-remove_unannotated_variant)
#   - update the format of multiallele tag
# 0.3.8 (2020.02.25)
#   - debugged first variant missing in annot module
# 0.3.7
#   - update annot module
# 0.3.6
#   - add makedata  
# 0.3.5
#   - merge some dbNSFP fields into transcript table (dbNSFPTranscript)
# 0.3.4
#   - change type of `dbNSFP SiPhy_29way_pi` to list
# 0.3.3
#   - encode 'space' to '%20' (remove blank space)


from . import stat
from . import gene
from . import variant
from .util import web

# Create your views here.


def home(request):
    data = {}
    return web.render(request, 'index.html', data)


def gene_list(request):
    data = {}
    return web.render(request, 'gene_list.html', data)


def gene_view(request):
    data = {}
    return web.render(request, 'gene_view.html', data)


def variant_list(request):
    data = {}
    return web.render(request, 'variant_list.html', data)


def variant_view(request):
    data = {}
    return web.render(request, 'variant_view.html', data)


def stat_view(request):
    data = {}
    return web.render(request, 'stat_list.html', data)





class Region:
    def __init__(self, region_str=""):
        self.region_str = region_str
        self.chrom = ""
        self.spos = 0
        self.epos = 0
        self.pars_region_str()

    def __str__(self):
        return self.chrom + ':' + str(self.spos) + '-' + str(self.epos)

    def pars_region_str(self):
        r = None
        if self.region_str != "":
            arr = self.region_str.split(':')
            arr2 = arr[1].split('-')
            r = {}
            self.chrom = arr[0].replace('chr', '')
            self.spos = int(arr2[0])
            self.epos = int(arr2[1])
    
    def get_split_region(self, bin_size=10000):
        regions = []
        for k in range(self.spos, self.epos, bin_size):
            r1 = Region()
            r1.chrom = self.chrom
            r1.spos = k
            r1.epos = r1.spos + bin_size - 1
            if r1.epos > self.epos:
                r1.epos = self.epos
            regions.append(r1)
        return regions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
import tabix
from .util import file_util
from .util import vcf_util
from .util import proc_util


def conv_cgap_filed(f1):
    return f1.replace('+', '').replace('-', '_')
    # return f1.replace('.','_')


class CONVFILE():
    def __init__(self, opt):
        self.opt = opt
        self.datastruct = file_util.load_json(opt['ds'])
        print("Staring... convert.CONVFILE")

    def get_header(self):
        rst = self.get_pars_cont(self.datastruct)
        return '#' + '\t'.join(rst) + "\n"

    def convert(self):
        print("Converting..")
        header = self.get_header()

        if self.opt['chromsplit']:
            fp = {}
            fname = {}
        else:
            fname = self.opt['out'] + '.tsv'
            fpc = open(fname, 'w')
            fpc.write(header)

        i = 0
        prev_chrom = ""

        if 'region' in self.opt.keys() and self.opt['region'] != '':
            tb = tabix.open(self.opt['in'])
            for arr in tb.querys(self.opt['region']):
                i += 1
                line = '\t'.join(arr)
                rst = self.get_pars_cont(self.datastruct, line, [])
                chrom = rst[0]
                if self.opt['chromsplit']:
                    if prev_chrom != chrom:
                        fname[chrom] = self.opt['out'] + '.' + chrom + '.tsv'
                        try:
                            fpc.close()
                        except AttributeError:
                            pass
                        fpc = open(fname[chrom], 'w')
                        fpc.write(header)
                        print("Saving...", fname[chrom])
                fpc.write('\t'.join(rst) + '\n')
                if i % 1000000 == 0:
                    print("...", rst[:2])
                prev_chrom = chrom
        else:
            for line in file_util.gzopen(self.opt['in']):
                if self.opt['in'].endswith('.vcf.gz'):
                    line = line.decode('UTF-8')
                if line[0] != '#':
                    i += 1
                    rst = self.get_pars_cont(self.datastruct, line, [])
                    chrom = rst[0]
                    if self.opt['chromsplit']:
                        if prev_chrom != chrom:
                            fname[chrom] = self.opt['out'] + '.' + chrom + '.tsv'
                            try:
                                fpc.close()
                            except AttributeError:
                                pass
                            fpc = open(fname[chrom], 'w')
                            fpc.write(header)
                            print("Saving...", fname[chrom])
                    fpc.write('\t'.join(rst) + '\n')
                    if i % 1000000 == 0:
                        print("...", rst[:2])
                    prev_chrom = chrom
        fpc.close()

        # TODO: convert this codes using tabix gzip library.
        time.sleep(60)
        if self.opt['chromsplit']:
            for chrom in fp.keys():
                cmd = "tabixgz " + fname[chrom]
                proc_util.run_cmd(cmd, True)
        else:
            cmd = "tabixgz " + fname
            proc_util.run_cmd(cmd, True)

    def conv_map_from_keyvar(self, arrcont):
        m = {}
        for cont in arrcont:
            arr = cont.strip().split('=')
            m[arr[0].strip()] = arr[1].strip()
        return m

    def get_pars_cont(self, ds, cont='', rst=[]):
        if cont != '':
            arrcont = cont.split(ds['delimiter'])
            arrcont[-1] = arrcont[-1].strip()
        for field_ds in ds['fields']:
            if 'fields' not in field_ds.keys():
                if cont != '':
                    if "type" in ds.keys() and ds['type'] == "keyvar":
                        arrcontmap = self.conv_map_from_keyvar(arrcont)
                        if field_ds['name'] in arrcontmap.keys():
                            rst.append(arrcontmap[field_ds['name']].strip())
                        else:
                            rst.append('')
                    else:
                        rst.append(arrcont[field_ds['colidx']].strip())
                else:
                    rst.append(field_ds['name'])
            else:
                if cont != '':
                    rst = self.get_pars_cont(field_ds, arrcont[field_ds['colidx']], rst)
                else:
                    rst = self.get_pars_cont(field_ds, '', rst)
        return rst


class VCF2TSV():
    vcfheader = None

    def __init__(self, opt):
        self.opt = opt
        self.vcfheader = None
        self.datastruct = file_util.load_json(opt['ds'])
        print('Staring... convert.VCF2TSV')

    def pars_vcfline(self, vcfline):
        v1 = vcf_util.pars_vcfline(vcfline, self.vcfheader.collist)
        for infokey in v1['INFO'].keys():
            if infokey in self.vcfheader.mutannofield.keys():
                arrd = []
                for f1 in v1['INFO'][infokey].split(','):
                    d = {}
                    arr = f1.split('|')
                    for i in range(len(self.vcfheader.mutannofield[infokey]['fieldlist'])):
                        d[self.vcfheader.mutannofield[infokey]['fieldlist'][i]] = arr[i].strip()
                    arrd.append(d)
                v1['INFO'][infokey] = arrd
        return v1

    def convert(self):
        out = self.opt['out']
        f = open(out, 'w')

        headercont = ""
        for line in file_util.gzopen(self.opt['in']):
            if self.opt['in'].endswith('.vcf.gz'):
                line = line.decode('UTF-8')

            if line[0] == "#":
                headercont += line
            else:
                if self.vcfheader is None:
                    self.vcfheader = vcf_util.VCFHEADER(headercont)

                v1 = self.pars_vcfline(line)
                # print (v1)
                cont = "### " + v1['CHROM'] + ":" + v1['POS'] + " " + v1['REF'] + ">" + v1['ALT'] + "\n"
                for source in self.vcfheader.mutannofield.keys():
                    if source != "MUTANNO":
                        cont += source + ' '
                        if source in v1['INFO'].keys():
                            cont += "(" + str(len(v1['INFO'][source])) + " fields)\n"
                            for f1 in v1['INFO'][source]:
                                cont += '-----------\n'
                                for fname in f1.keys():
                                    cont += "\t" + fname + " : " + f1[fname] + '\n'
                        else:
                            cont += "(0 fields)\n"

                # print (cont)
                f.write(cont)

                '''
                valuemap = {}
                for field in infofields:
                    if '=' in field:
                        arrf = field.split('=')
                        k1 = arrf[0]
                        if k1 in fieldnames.keys():
                            values = arrf[1].split('|')
                            print (k1, len(values), len(fieldnames[k1])),
                            for k in range(len(fieldnames[k1])):
                                valuemap[k1+"_"+fieldnames[k1][k]] = values[k]
                # print (arr[7])
                valuelist = []
                for k1 in fieldkeylist:
                    for fidx in range(len(fieldnames[k1])):
                        print (k1, fieldnames[k1][fidx])
                        try:
                            valuelist.append( decode_infovalue(valuemap[k1+"_"+fieldnames[k1][fidx]]) )
                        except KeyError:
                            valuelist.append('')
                f.write('\t'.join(valuelist)+'\n')
                '''
        # print (fieldnames)
        # print (colnames)
        f.close()
        print('Saved', out)


class VEP2TAB():
    def __init__(self, opt):
        self.opt = opt
        print('Staring... convert.VEP2TAB')

    def convert(self):
        out = self.opt['out']
        f = open(out, 'w')

        for line in file_util.gzopen(self.opt['in']):
            if self.opt['in'].endswith('.gz'):
                line = line.decode('UTF-8')
            arr = line.split('\t')
            if line[0] == "#":
                if arr[0] == "#Uploaded_variation":
                    cont = ['#CHROM', 'POS', 'REF', 'ALT']
                    arr[0] = arr[0][1:]
                    cont.extend(arr)
                    f.write('\t'.join(cont))
            else:
                arr2 = arr[0].split('_')
                cont = [arr2[0].replace('chr', ''), arr2[1], arr2[2].replace('/', '\t')]
                cont.extend(arr)
                f.write('\t'.join(cont))
        f.close()
        print('Saved ' + out)
        # cmd = "bgzip -c " + out + " > " + out + ".gz;tabix -f -p vcf " + out + ".gz;"
        # proc_util.run_cmd(cmd, True)


if __name__ == "__main__":
    pass



#!/usr/bin/env python
import os
import sys
import subprocess
from io import StringIO


def cpu_num():
    import psutil
    return psutil.NUM_CPUS


def mem_usage():
    import psutil
    return psutil.virtual_memory()


def cpu_usage():
    import psutil
    return psutil.cpu_percent(interval=1, percpu=False)


def disk_usage(path="/"):
    import psutil
    return psutil.disk_usage(path)


def back_run(scmd):
    import file_util
    svrname = os.uname()[1]
    fname = "./tmp_" + svrname + ".sh"
    #file_util.fileSave (fname, scmd.strip() + " & \n", "w")
    file_util.fileSave(fname, scmd.strip() + " > /dev/null 2>&1", "w")
    run_cmd("chmod 755 " + fname)
    run_cmd(fname + " & ")


def run_cmd(scmd, flag=False):
    
    if flag:
        print(scmd)
    rst = os.popen(scmd)
    rst_cont = rst.read()
    # rst = subprocess.Popen([scmd], stdout=subprocess.PIPE, shell=True)
    # (rst_cont, error) = rst.communicate()
    # sys.stdout = old_stdout

    return rst_cont


def proc_list(v=""):
    import str_util
    cmd = "ps -ef"
    if v != "":
        cmd += " | grep " + v
    cont = run_cmd(cmd)
    arr = cont.strip().split("\n")
    plist = []
    for line in arr:
        p = {}
        arr2 = str_util.strip_space(line).split(" ")
        p['user'] = arr2[0]
        p['pid'] = arr2[1]
        p['pid2'] = arr2[2]
        p['id'] = arr2[3]
        p['start_time'] = arr2[4]
        p['job_id'] = arr2[5]
        p['run_time'] = arr2[6]
        p['proc'] = " ".join(arr2[7:])
        plist.append(p)

    return plist


def top_list():
    cmd = "top -b -n 1"
    cont = run_cmd(cmd)
    arr_line = cont.strip().split("\n")
    flag = False
    rst = {}
    rst['joblist'] = []
    for line in arr_line:
        if "Cpu(s):" in line:
            line = line.strip().replace("  ", " ").replace("  ", " ").replace("  ", " ")
            line = line.replace("  ", " ").replace("  ", " ").replace("  ", " ")
            arr = line.split(" ")
            rst['cpu_usage'] = arr[1].replace("%us,", "")
        if "Mem:" in line:
            line = line.strip().replace("  ", " ").replace("  ", " ").replace("  ", " ")
            line = line.replace("  ", " ").replace("  ", " ").replace("  ", " ")
            arr = line.split(" ")
            rst['mem_total'] = arr[1]
            rst['mem_used'] = arr[2]
            rst['mem_free'] = arr[3]
        if flag:
            line = line.strip().replace("  ", " ").replace("  ", " ").replace("  ", " ")
            line = line.replace("  ", " ").replace("  ", " ").replace("  ", " ")
            arr = line.split(" ")
            if arr[8] == "0":
                break
            m = {}
            m['PID'] = arr[0]
            m['USER'] = arr[1]
            m['CPU'] = arr[8]
            m['MEM'] = arr[9]
            m['TIME'] = arr[10]
            m['COMMAND'] = arr[11]
            rst['joblist'].append(m)

        if "USER" in line:
            line = line.replace("  ", " ").replace("  ", " ").replace("  ", " ")
            line = line.replace("  ", " ").replace("  ", " ").replace("  ", " ")
            header = line.split(" ")
            flag = True
        # print line

    rst['no_working_job'] = len(rst['joblist'])
    return rst


def in_proc_list(proc_name):
    plist = proc_list()
    flag = False
    for p in plist:
        if proc_name in p['proc']:
            flag = True
            break
    return flag


def cnt_proc(proc_name):
    plist = proc_list()
    flag = False
    cnt = 0
    for p in plist:
        if proc_name in p['proc']:
            cnt += 1
            break
    return cnt

# For job scheduling system in orchestra


def get_bjob_list():
    cmd = "bjobs -aw"
    rst = run_cmd(cmd)
    bsubjoblist = []
    i = 0
    for line in rst.split("\n"):
        # print line
        if i > 0 and len(line) > 20:
            for kk in range(10):
                line = line.replace("  ", " ")
            arr = line.split(" ")
            bjob = {}
            bjob['id'] = arr[0]
            bjob['status'] = arr[2]
            bjob['sh'] = " ".join(arr[6:])
            bsubjoblist.append(bjob)
        i += 1
    return bsubjoblist

# For job scheduling system in orchestra


def get_bjob(tag, bjob_list=[]):
    if len(bjob_list) == 0:
        bjob_list = get_bjob_list()
    tjob = {'status': 'NA', 'id': ''}
    for bjob in bjob_list:
        if tag in bjob['sh']:
            tjob = bjob
            break
    return tjob


'''
# alternative approach
def run_cmd (u_cmd):
	p = Popen([u_cmd], shell=True, stdout=PIPE)
	#p.wait()
	return p.stdout.read()
'''


def send_googletalk(msg, to="intellims@gmail.com"):
    scmd = "echo 'msg:"+msg+"' | sendxmpp -f /home/pcaso/.sendxmpprc -t -u minseok.d.kwon -o gmail.com -p qkdehdk100 " + to
    run_cmd(scmd)


class RemoteSvr():

    def __init__(self, server_name=""):
        if server_name == "":
            import os
            self.server_name = os.uname()[1]
        else:
            self.server_name = server_name

    def update_status(self, path=""):
        import web_util
        import db_util
        import file_util
        self.logpath = path
        d = {}
        d['cpu_usage'] = cpu_usage()
        d['server_name'] = self.server_name
        d['check_time'] = "NOW()"
        sql = db_util.insertSQL(d, "server_status")
        if path == "":
            web_util.dbgate(sql, "pcaso")
        else:
            import time
            file_util.fileSave(path + "/" + self.server_name + ".log", str(time.time()) + "\t" + sql+"\n", "w")

    def log_updater(self):
        import file_util
        import web_util

        backupkey = []
        for line in open(self.logpath + "/prev.log"):
            backupkey.append(line.strip().split("\t")[0])

        cont = ""
        for fname in file_util.walk(self.logpath):
            for line in open(fname):
                arr = line.strip().split("\t")
                if not arr[0] in backupkey:
                    web_util.dbgate(arr[1], "pcaso")
                    cont += line

        file_util.fileSave(self.logpath + "/prev.log", cont, "w")

    def job_run(self, jobfile):
        import file_util
        if file_util.is_exist(jobfile) and cpu_usage() < 90:
            i = 0
            cont = ""
            scmd = ""
            for line in open(jobfile):
                if i == 0:
                    scmd = line.strip()
                else:
                    cont += line
                i += 1
            file_util.fileSave(jobfile, cont, "w")
            if len(scmd) > 0:
                back_run(scmd)
                if i > 1:
                    self.job_run(jobfile)


if __name__ == "__main__":
    run_cmd("ls")



#!/usr/bin/env python
from setuptools import setup, find_packages

install_requires = [
    'pyfaidx>=0.5.3.1',
    'pysam>=0.11.2.2y7',
    'pytabix>=0.0.2',
    'pyliftover>=0.4',
    'xmltodict>=0.12.0',
    'pyvcf>=0.6.8'
]

tests_require = [
    'nose',
    'mock'
]

extras_require = {
    'docs': [
        'Sphinx>=1.1'
    ]
}


setup(name='mutanno',
      version='0.4.1',
      url='https://github.com/dbmi-bgm/mutanno',
      license='MIT',
      author='CGAP team at Harvard Medical School',
      author_email='daniel.minseok.kwon@gmail.com',
      description='Mutation Annotation Tool',
      download_url='https://github.com/dbmi-bgm/mutanno/archive/0.4.0.tar.gz',
      keywords=['genomics', 'bioinformatics'],
      classifiers=[
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Libraries',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
      ],
      # packages=find_packages(exclude=['tests']),
      python_requires=">=3.4",
      packages=find_packages('src'),
      package_dir={'': 'src'},
      package_data={'mutanno.web.templates': ['*.html'],
                    'mutanno.web.static.css': ['*.css'],
                    'mutanno.web.static.fonts': ['*.woff', '*.woff2'],
                    'mutanno.web.static.img': ['*.png', '*.gif'],
                    'mutanno.web.static.js': ['*.js', '*.png', '*.map'],
                    'mutanno.web.static.semantic': ['*.js', '*.json', '*.css', '*.md', 'LICENSE'],
                    'mutanno.web.static.semantic.components': ['*.js', '*.css'],
                    'mutanno.web.static.semantic.themes.default.assets.images': ['*.png'],
                    'mutanno.web.static.semantic.themes.default.assets.fonts': ['*.eot', '*.svg', '*.ttf', '*.woff', '*.woff2', '*.otf']
                    },
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      zip_safe=False,
      install_requires=install_requires,
      setup_requires=['nose>=1.0'],
      test_suite='nose.collector',
      entry_points={
          'console_scripts': [
              'mutanno=mutanno:cli',
          ]
      })
